# Copyright 2019 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from unittest import mock

from neutron_lib import constants as const

from neutron.common.ovn import acl
from neutron.common.ovn import constants as ovn_const
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import impl_idl_ovn
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovn_client
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovn_db_sync
from neutron.services.ovn_l3 import plugin as ovn_plugin
from neutron.tests.unit import fake_resources as fakes
from neutron.tests.unit.plugins.ml2.drivers.ovn.mech_driver import \
    test_mech_driver


OvnPortInfo = collections.namedtuple('OvnPortInfo', ['name'])


@mock.patch.object(ovn_plugin.OVNL3RouterPlugin, '_sb_ovn', mock.Mock())
class TestOvnNbSyncML2(test_mech_driver.OVNMechanismDriverTestCase):

    l3_plugin = 'ovn-router'

    def setUp(self):
        super(TestOvnNbSyncML2, self).setUp()

        self.subnet = {'cidr': '10.0.0.0/24',
                       'id': 'subnet1',
                       'subnetpool_id': None,
                       'name': 'private-subnet',
                       'enable_dhcp': True,
                       'network_id': 'n1',
                       'tenant_id': 'tenant1',
                       'gateway_ip': '10.0.0.1',
                       'ip_version': 4,
                       'shared': False}
        self.matches = ["", "", "", ""]

        self.networks = [{'id': 'n1',
                          'mtu': 1450,
                          'provider:physical_network': 'physnet1',
                          'provider:segmentation_id': 1000},
                         {'id': 'n2',
                          'mtu': 1450},
                         {'id': 'n4',
                          'mtu': 1450,
                          'provider:physical_network': 'physnet2'}]

        self.segments = [{'id': 'seg1',
                          'network_id': 'n1',
                          'physical_network': 'physnet1',
                          'network_type': 'vlan',
                          'segmentation_id': 1000},
                         {'id': 'seg2',
                          'network_id': 'n2',
                          'physical_network': None,
                          'network_type': 'geneve'},
                         {'id': 'seg4',
                          'network_id': 'n4',
                          'physical_network': 'physnet2',
                          'network_type': 'flat'}]
        self.segments_map = {
            k['network_id']: k
            for k in self.segments}

        self.subnets = [{'id': 'n1-s1',
                         'network_id': 'n1',
                         'enable_dhcp': True,
                         'cidr': '10.0.0.0/24',
                         'tenant_id': 'tenant1',
                         'gateway_ip': '10.0.0.1',
                         'dns_nameservers': [],
                         'host_routes': [],
                         'ip_version': 4},
                        {'id': 'n1-s2',
                         'network_id': 'n1',
                         'enable_dhcp': True,
                         'cidr': 'fd79:e1c:a55::/64',
                         'tenant_id': 'tenant1',
                         'gateway_ip': 'fd79:e1c:a55::1',
                         'dns_nameservers': [],
                         'host_routes': [],
                         'ip_version': 6},
                        {'id': 'n2',
                         'network_id': 'n2',
                         'enable_dhcp': True,
                         'cidr': '20.0.0.0/24',
                         'tenant_id': 'tenant1',
                         'gateway_ip': '20.0.0.1',
                         'dns_nameservers': [],
                         'host_routes': [],
                         'ip_version': 4}]

        self.security_groups = [
            {'id': 'sg1', 'tenant_id': 'tenant1',
             'security_group_rules': [{'remote_group_id': None,
                                       'direction': 'ingress',
                                       'remote_ip_prefix': '0.0.0.0/0',
                                       'protocol': 'tcp',
                                       'ethertype': 'IPv4',
                                       'tenant_id': 'tenant1',
                                       'port_range_max': 65535,
                                       'port_range_min': 1,
                                       'id': 'ruleid1',
                                       'security_group_id': 'sg1'}],
             'name': 'all-tcp'},
            {'id': 'sg2', 'tenant_id': 'tenant1',
             'security_group_rules': [{'remote_group_id': 'sg2',
                                       'direction': 'egress',
                                       'remote_ip_prefix': '0.0.0.0/0',
                                       'protocol': 'tcp',
                                       'ethertype': 'IPv4',
                                       'tenant_id': 'tenant1',
                                       'port_range_max': 65535,
                                       'port_range_min': 1,
                                       'id': 'ruleid1',
                                       'security_group_id': 'sg2'}],
             'name': 'all-tcpe'}]

        self.sg_port_groups_ovn = [mock.Mock(), mock.Mock(), mock.Mock()]
        self.sg_port_groups_ovn[0].configure_mock(
            name='pg_sg1',
            external_ids={ovn_const.OVN_SG_EXT_ID_KEY: 'sg1'},
            ports=[],
            acls=[])
        self.sg_port_groups_ovn[1].configure_mock(
            name='pg_unknown_del',
            external_ids={ovn_const.OVN_SG_EXT_ID_KEY: 'sg2'},
            ports=[],
            acls=[])
        self.sg_port_groups_ovn[2].configure_mock(
            name='neutron_pg_drop',
            external_ids=[],
            ports=[],
            acls=[])

        self.ports = [
            {'id': 'p1n1',
             'device_owner': 'compute:None',
             'fixed_ips':
                 [{'subnet_id': 'b142f5e3-d434-4740-8e88-75e8e5322a40',
                   'ip_address': '10.0.0.4'},
                  {'subnet_id': 'subnet1',
                   'ip_address': 'fd79:e1c:a55::816:eff:eff:ff2'}],
             'security_groups': ['sg1'],
             'network_id': 'n1'},
            {'id': 'p2n1',
             'device_owner': 'compute:None',
             'fixed_ips':
                 [{'subnet_id': 'b142f5e3-d434-4740-8e88-75e8e5322a40',
                   'ip_address': '10.0.0.4'},
                  {'subnet_id': 'subnet1',
                   'ip_address': 'fd79:e1c:a55::816:eff:eff:ff2'}],
             'security_groups': ['sg2'],
             'network_id': 'n1',
             'extra_dhcp_opts': [{'ip_version': 6,
                                  'opt_name': 'domain-search',
                                  'opt_value': 'foo-domain'}]},
            {'id': 'p1n2',
             'device_owner': 'compute:None',
             'fixed_ips':
                 [{'subnet_id': 'b142f5e3-d434-4740-8e88-75e8e5322a40',
                   'ip_address': '10.0.0.4'},
                  {'subnet_id': 'subnet1',
                   'ip_address': 'fd79:e1c:a55::816:eff:eff:ff2'}],
             'security_groups': ['sg1'],
             'network_id': 'n2',
             'extra_dhcp_opts': [{'ip_version': 4,
                                  'opt_name': 'tftp-server',
                                  'opt_value': '20.0.0.20'},
                                 {'ip_version': 4,
                                  'opt_name': 'dns-server',
                                  'opt_value': '8.8.8.8'},
                                 {'ip_version': 6,
                                  'opt_name': 'domain-search',
                                  'opt_value': 'foo-domain'}]},
            {'id': 'p2n2',
             'device_owner': 'compute:None',
             'fixed_ips':
                 [{'subnet_id': 'b142f5e3-d434-4740-8e88-75e8e5322a40',
                   'ip_address': '10.0.0.4'},
                  {'subnet_id': 'subnet1',
                   'ip_address': 'fd79:e1c:a55::816:eff:eff:ff2'}],
             'security_groups': ['sg2'],
             'network_id': 'n2'},
            {'id': 'fp1',
             'device_owner': 'network:floatingip',
             'fixed_ips':
                 [{'subnet_id': 'ext-subnet',
                   'ip_address': '90.0.0.10'}],
             'network_id': 'ext-net'}]

        self.ports_ovn = [OvnPortInfo('p1n1'), OvnPortInfo('p1n2'),
                          OvnPortInfo('p2n1'), OvnPortInfo('p2n2'),
                          OvnPortInfo('p3n1'), OvnPortInfo('p3n3')]

        self.acls_ovn = {
            'lport1':
            # ACLs need to be removed by the sync tool
            [{'id': 'acl1', 'priority': 00, 'policy': 'allow',
              'lswitch': 'lswitch1', 'lport': 'lport1'}],
            'lport2':
            [{'id': 'acl2', 'priority': 00, 'policy': 'drop',
             'lswitch': 'lswitch2', 'lport': 'lport2'}],
            # ACLs need to be kept as-is by the sync tool
            'p2n2':
            [{'lport': 'p2n2', 'direction': 'to-lport',
              'log': False, 'lswitch': 'neutron-n2',
              'priority': 1001, 'action': 'drop',
             'external_ids': {'neutron:lport': 'p2n2'},
              'match': 'outport == "p2n2" && ip'},
             {'lport': 'p2n2', 'direction': 'to-lport',
              'log': False, 'lswitch': 'neutron-n2',
              'priority': 1002, 'action': 'allow',
              'external_ids': {'neutron:lport': 'p2n2'},
              'match': 'outport == "p2n2" && ip4 && '
              'ip4.src == 10.0.0.0/24 && udp && '
              'udp.src == 67 && udp.dst == 68'}]}

        self.routers = [{'id': 'r1', 'routes': [{'nexthop': '20.0.0.100',
                         'destination': '11.0.0.0/24'}, {
                         'nexthop': '20.0.0.101',
                         'destination': '12.0.0.0/24',
                         'external_ids': {}}],
                         'gw_port_id': 'gpr1',
                         'external_gateway_info': {
                             'network_id': "ext-net", 'enable_snat': True,
                             'external_fixed_ips': [
                                 {'subnet_id': 'ext-subnet',
                                  'ip_address': '90.0.0.2'}]}},
                        {'id': 'r2', 'routes': [{'nexthop': '40.0.0.100',
                         'destination': '30.0.0.0/24',
                         'external_ids': {}}],
                         'gw_port_id': 'gpr2',
                         'external_gateway_info': {
                             'network_id': "ext-net", 'enable_snat': True,
                             'external_fixed_ips': [
                                 {'subnet_id': 'ext-subnet',
                                  'ip_address': '100.0.0.2'}]}},
                        {'id': 'r4', 'routes': []}]

        self.get_sync_router_ports = [
            {'fixed_ips': [{'subnet_id': 'subnet1',
                            'ip_address': '192.168.1.1'}],
             'id': 'p1r1',
             'device_id': 'r1',
             'mac_address': 'fa:16:3e:d7:fd:5f'},
            {'fixed_ips': [{'subnet_id': 'subnet2',
                            'ip_address': '192.168.2.1'}],
             'id': 'p1r2',
             'device_id': 'r2',
             'mac_address': 'fa:16:3e:d6:8b:ce'},
            {'fixed_ips': [{'subnet_id': 'subnet4',
                            'ip_address': '192.168.4.1'}],
             'id': 'p1r4',
             'device_id': 'r4',
             'mac_address': 'fa:16:3e:12:34:56'}]

        self.floating_ips = [{'id': 'fip1', 'router_id': 'r1',
                              'floating_ip_address': '90.0.0.10',
                              'fixed_ip_address': '172.16.0.10'},
                             {'id': 'fip2', 'router_id': 'r1',
                              'floating_ip_address': '90.0.0.12',
                              'fixed_ip_address': '172.16.2.12'},
                             {'id': 'fip3', 'router_id': 'r2',
                              'floating_ip_address': '100.0.0.10',
                              'fixed_ip_address': '192.168.2.10'},
                             {'id': 'fip4', 'router_id': 'r2',
                              'floating_ip_address': '100.0.0.11',
                              'fixed_ip_address': '192.168.2.11'}]

        self.lrouters_with_rports = [{'name': 'r3',
                                      'ports': {'p1r3': ['fake']},
                                      'static_routes': [],
                                      'snats': [],
                                      'dnat_and_snats': []},
                                     {'name': 'r4',
                                      'ports': {'p1r4':
                                                ['fdad:123:456::1/64',
                                                 'fdad:789:abc::1/64']},
                                      'static_routes': [],
                                      'snats': [],
                                      'dnat_and_snats': []},
                                     {'name': 'r1',
                                      'ports': {'p3r1': ['fake']},
                                      'static_routes':
                                      [{'nexthop': '20.0.0.100',
                                        'destination': '11.0.0.0/24'},
                                       {'nexthop': '20.0.0.100',
                                        'destination': '10.0.0.0/24'}],
                                      'snats':
                                      [{'logical_ip': '172.16.0.0/24',
                                        'external_ip': '90.0.0.2',
                                        'type': 'snat'},
                                       {'logical_ip': '172.16.1.0/24',
                                        'external_ip': '90.0.0.2',
                                        'type': 'snat'}],
                                      'dnat_and_snats':
                                      [{'logical_ip': '172.16.0.10',
                                        'external_ip': '90.0.0.10',
                                        'type': 'dnat_and_snat'},
                                       {'logical_ip': '172.16.1.11',
                                        'external_ip': '90.0.0.11',
                                        'type': 'dnat_and_snat'},
                                       {'logical_ip': '192.168.2.11',
                                        'external_ip': '100.0.0.11',
                                        'type': 'dnat_and_snat',
                                        'external_mac': '01:02:03:04:05:06',
                                        'logical_port': 'vm1'}]}]

        self.lswitches_with_ports = [{'name': 'neutron-n1',
                                      'ports': ['p1n1', 'p3n1'],
                                      'provnet_ports': []},
                                     {'name': 'neutron-n3',
                                      'ports': ['p1n3', 'p2n3'],
                                      'provnet_ports': []},
                                     {'name': 'neutron-n4',
                                      'ports': [],
                                      'provnet_ports': [
                                          'provnet-seg4',
                                          'provnet-orphaned-segment']}]

        self.lrport_networks = ['fdad:123:456::1/64', 'fdad:cafe:a1b2::1/64']

    def get_additional_service_plugins(self):
        p = super(TestOvnNbSyncML2, self).get_additional_service_plugins()
        p.update({'segments': 'neutron.services.segments.plugin.Plugin'})
        return p

    def _fake_get_ovn_dhcp_options(self, subnet, network, server_mac=None):
        if subnet['id'] == 'n1-s1':
            return {'cidr': '10.0.0.0/24',
                    'options': {'server_id': '10.0.0.1',
                                'server_mac': '01:02:03:04:05:06',
                                'lease_time': str(12 * 60 * 60),
                                'mtu': '1450',
                                'router': '10.0.0.1'},
                    'external_ids': {'subnet_id': 'n1-s1'}}
        return {'cidr': '', 'options': '', 'external_ids': {}}

    def _fake_get_gw_info(self, ctx, router):
        return {
            'r1': [ovn_client.GW_INFO(router_ip='90.0.0.2',
                                      gateway_ip='90.0.0.1',
                                      network_id='', subnet_id='ext-subnet',
                                      ip_version=4,
                                      ip_prefix=const.IPv4_ANY)],
            'r2': [ovn_client.GW_INFO(router_ip='100.0.0.2',
                                      gateway_ip='100.0.0.1',
                                      network_id='', subnet_id='ext-subnet',
                                      ip_version=4,
                                      ip_prefix=const.IPv4_ANY)]
        }.get(router['id'], [])

    def _fake_get_v4_network_of_all_router_ports(self, ctx, router_id):
        return {'r1': ['172.16.0.0/24', '172.16.2.0/24'],
                'r2': ['192.168.2.0/24']}.get(router_id, [])

    def _test_mocks_helper(self, ovn_nb_synchronizer):
        core_plugin = ovn_nb_synchronizer.core_plugin
        ovn_api = ovn_nb_synchronizer.ovn_api
        ovn_driver = ovn_nb_synchronizer.ovn_driver
        l3_plugin = ovn_nb_synchronizer.l3_plugin
        pf_plugin = ovn_nb_synchronizer.pf_plugin
        segments_plugin = ovn_nb_synchronizer.segments_plugin

        core_plugin.get_networks = mock.Mock()
        core_plugin.get_networks.return_value = self.networks
        core_plugin.get_subnets = mock.Mock()
        core_plugin.get_subnets.return_value = self.subnets

        def get_segments(self, filters):
            segs = []
            for segment in self.segments:
                if segment['network_id'] == filters['network_id'][0]:
                    segs.append(segment)
            return segs

        segments_plugin.get_segments = mock.Mock()
        segments_plugin.get_segments.side_effect = (
            lambda x, filters: get_segments(self, filters))

        # following block is used for acl syncing unit-test

        # With the given set of values in the unit testing,
        # 19 neutron acls should have been there,
        # 4 acls are returned as current ovn acls,
        # two of which will match with neutron.
        # So, in this example 17 will be added, 2 removed
        core_plugin.get_ports = mock.Mock()
        core_plugin.get_ports.return_value = self.ports
        mock.patch.object(acl, '_get_subnet_from_cache',
                          return_value=self.subnet).start()
        mock.patch.object(acl, 'acl_remote_group_id',
                          side_effect=self.matches).start()
        core_plugin.get_security_group = mock.MagicMock(
            side_effect=self.security_groups)
        ovn_nb_synchronizer.get_acls = mock.Mock()
        ovn_nb_synchronizer.get_acls.return_value = self.acls_ovn
        core_plugin.get_security_groups = mock.MagicMock(
            return_value=self.security_groups)
        get_sg_port_groups = mock.MagicMock()
        get_sg_port_groups.execute.return_value = self.sg_port_groups_ovn
        ovn_api.db_list_rows.return_value = get_sg_port_groups
        ovn_api.lsp_list.execute.return_value = self.ports_ovn
        # end of acl-sync block

        # The following block is used for router and router port syncing tests
        # With the give set of values in the unit test,
        # The Neutron db has Routers r1 and r2 present.
        # The OVN db has Routers r1 and r3 present.
        # During the sync r2 will need to be created and r3 will need
        # to be deleted from the OVN db. When Router r3 is deleted, all LRouter
        # ports associated with r3 is deleted too.
        #
        # Neutron db has Router ports p1r1 in Router r1 and p1r2 in Router r2
        # OVN db has p1r3 in Router 3 and p3r1 in Router 1.
        # During the sync p1r1 and p1r2 will be added and p1r3 and p3r1
        # will be deleted from the OVN db
        l3_plugin.get_routers = mock.Mock()
        l3_plugin.get_routers.return_value = self.routers
        l3_plugin._get_sync_interfaces = mock.Mock()
        l3_plugin._get_sync_interfaces.return_value = (
            self.get_sync_router_ports)
        ovn_nb_synchronizer._ovn_client = mock.Mock()
        ovn_nb_synchronizer._ovn_client.\
            _get_nets_and_ipv6_ra_confs_for_router_port.return_value = (
                self.lrport_networks, {'fixed_ips': {}})
        ovn_nb_synchronizer._ovn_client._get_v4_network_of_all_router_ports. \
            side_effect = self._fake_get_v4_network_of_all_router_ports
        ovn_nb_synchronizer._ovn_client._get_gw_info = mock.Mock()
        ovn_nb_synchronizer._ovn_client._get_gw_info.side_effect = (
            self._fake_get_gw_info)
        # end of router-sync block
        l3_plugin.get_floatingips = mock.Mock()
        l3_plugin.get_floatingips.return_value = self.floating_ips
        pf_plugin.get_floatingip_port_forwardings = mock.Mock(return_value=[])
        ovn_api.get_all_logical_switches_with_ports = mock.Mock()
        ovn_api.get_all_logical_switches_with_ports.return_value = (
            self.lswitches_with_ports)

        ovn_api.get_all_logical_routers_with_rports = mock.Mock()
        ovn_api.get_all_logical_routers_with_rports.return_value = (
            self.lrouters_with_rports)

        ovn_api.transaction = mock.MagicMock()

        ovn_nb_synchronizer._ovn_client.create_network = mock.Mock()
        ovn_nb_synchronizer._ovn_client.create_port = mock.Mock()
        ovn_driver.validate_and_get_data_from_binding_profile = mock.Mock()
        ovn_nb_synchronizer._ovn_client.create_port = mock.Mock()
        ovn_nb_synchronizer._ovn_client.create_port.return_value = mock.ANY
        ovn_nb_synchronizer._ovn_client.create_provnet_port = mock.Mock()
        ovn_api.ls_del = mock.Mock()
        ovn_api.delete_lswitch_port = mock.Mock()

        ovn_api.delete_lrouter = mock.Mock()
        ovn_api.delete_lrouter_port = mock.Mock()
        ovn_api.add_static_route = mock.Mock()
        ovn_api.delete_static_route = mock.Mock()
        ovn_api.get_all_dhcp_options.return_value = {
            'subnets': {'n1-s1': {'cidr': '10.0.0.0/24',
                                  'options':
                                  {'server_id': '10.0.0.1',
                                   'server_mac': '01:02:03:04:05:06',
                                   'lease_time': str(12 * 60 * 60),
                                   'mtu': '1450',
                                   'router': '10.0.0.1'},
                                  'external_ids': {'subnet_id': 'n1-s1'},
                                  'uuid': 'UUID1'},
                        'n1-s3': {'cidr': '30.0.0.0/24',
                                  'options':
                                  {'server_id': '30.0.0.1',
                                   'server_mac': '01:02:03:04:05:06',
                                   'lease_time': str(12 * 60 * 60),
                                   'mtu': '1450',
                                   'router': '30.0.0.1'},
                                  'external_ids': {'subnet_id': 'n1-s3'},
                                  'uuid': 'UUID2'}},
            'ports_v4': {'p1n2': {'cidr': '10.0.0.0/24',
                                  'options': {'server_id': '10.0.0.1',
                                              'server_mac':
                                                  '01:02:03:04:05:06',
                                              'lease_time': '1000',
                                              'mtu': '1400',
                                              'router': '10.0.0.1'},
                                  'external_ids': {'subnet_id': 'n1-s1',
                                                   'port_id': 'p1n2'},
                                  'uuid': 'UUID3'},
                         'p5n2': {'cidr': '10.0.0.0/24',
                                  'options': {'server_id': '10.0.0.1',
                                              'server_mac':
                                                  '01:02:03:04:05:06',
                                              'lease_time': '1000',
                                              'mtu': '1400',
                                              'router': '10.0.0.1'},
                                  'external_ids': {'subnet_id': 'n1-s1',
                                                   'port_id': 'p5n2'},
                                  'uuid': 'UUID4'}},
            'ports_v6': {'p1n1': {'cidr': 'fd79:e1c:a55::/64',
                                  'options': {'server_id': '01:02:03:04:05:06',
                                              'mtu': '1450'},
                                  'external_ids': {'subnet_id': 'fake',
                                                   'port_id': 'p1n1'},
                                  'uuid': 'UUID5'},
                         'p1n2': {'cidr': 'fd79:e1c:a55::/64',
                                  'options': {'server_id': '01:02:03:04:05:06',
                                              'mtu': '1450'},
                                  'external_ids': {'subnet_id': 'fake',
                                                   'port_id': 'p1n2'},
                                  'uuid': 'UUID6'}}}

        ovn_nb_synchronizer._ovn_client._add_subnet_dhcp_options = mock.Mock()
        ovn_nb_synchronizer._ovn_client._get_ovn_dhcp_options = mock.Mock()
        ovn_nb_synchronizer._ovn_client._get_ovn_dhcp_options.side_effect = (
            self._fake_get_ovn_dhcp_options)
        ovn_api.delete_dhcp_options = mock.Mock()
        ovn_nb_synchronizer._ovn_client.get_port_dns_records = mock.Mock()
        ovn_nb_synchronizer._ovn_client.get_port_dns_records.return_value = {}

    def _test_ovn_nb_sync_helper(self, ovn_nb_synchronizer,
                                 networks, ports,
                                 routers, router_ports,
                                 create_router_list, create_router_port_list,
                                 update_router_port_list,
                                 del_router_list, del_router_port_list,
                                 create_network_list, create_port_list,
                                 create_provnet_port_list,
                                 del_network_list, del_port_list,
                                 add_static_route_list, del_static_route_list,
                                 add_snat_list, del_snat_list,
                                 add_floating_ip_list, del_floating_ip_list,
                                 add_subnet_dhcp_options_list,
                                 delete_dhcp_options_list,
                                 add_port_groups_list,
                                 del_port_groups_list):
        self._test_mocks_helper(ovn_nb_synchronizer)

        ovn_api = ovn_nb_synchronizer.ovn_api
        mock.patch.object(impl_idl_ovn.OvsdbNbOvnIdl, 'from_worker').start()

        ovn_nb_synchronizer.do_sync()

        create_port_groups_calls = [mock.call(**a)
                                    for a in add_port_groups_list]
        self.assertEqual(
            len(add_port_groups_list),
            ovn_api.pg_add.call_count)
        ovn_api.pg_add.assert_has_calls(
            create_port_groups_calls, any_order=True)

        del_port_groups_calls = [mock.call(d)
                                 for d in del_port_groups_list]
        self.assertEqual(
            len(del_port_groups_list),
            ovn_api.pg_del.call_count)
        ovn_api.pg_del.assert_has_calls(
            del_port_groups_calls, any_order=True)

        self.assertEqual(
            len(create_network_list),
            ovn_nb_synchronizer._ovn_client.create_network.call_count)
        create_network_calls = [mock.call(mock.ANY, net['net'])
                                for net in create_network_list]
        ovn_nb_synchronizer._ovn_client.create_network.assert_has_calls(
            create_network_calls, any_order=True)

        self.assertEqual(
            len(create_port_list),
            ovn_nb_synchronizer._ovn_client.create_port.call_count)
        create_port_calls = [mock.call(mock.ANY, port)
                             for port in create_port_list]
        ovn_nb_synchronizer._ovn_client.create_port.assert_has_calls(
            create_port_calls, any_order=True)

        create_provnet_port_calls = [
            mock.call(
                network['id'],
                self.segments_map[network['id']],
                txn=mock.ANY)
            for network in create_provnet_port_list
            if network.get('provider:physical_network')]
        self.assertEqual(
            len(create_provnet_port_list),
            ovn_nb_synchronizer._ovn_client.create_provnet_port.call_count)
        ovn_nb_synchronizer._ovn_client.create_provnet_port.assert_has_calls(
            create_provnet_port_calls, any_order=True)

        self.assertEqual(len(del_network_list),
                         ovn_api.ls_del.call_count)
        ls_del_calls = [mock.call(net_name)
                        for net_name in del_network_list]
        ovn_api.ls_del.assert_has_calls(
            ls_del_calls, any_order=True)

        self.assertEqual(len(del_port_list),
                         ovn_api.delete_lswitch_port.call_count)
        delete_lswitch_port_calls = [mock.call(lport_name=port['id'],
                                               lswitch_name=port['lswitch'])
                                     for port in del_port_list]
        ovn_api.delete_lswitch_port.assert_has_calls(
            delete_lswitch_port_calls, any_order=True)

        add_route_calls = [mock.call(mock.ANY, ip_prefix=route['destination'],
                                     nexthop=route['nexthop'],
                                     external_ids=route.get('external_ids',
                                     {}))
                           for route in add_static_route_list]
        ovn_api.add_static_route.assert_has_calls(add_route_calls,
                                                  any_order=True)
        self.assertEqual(len(add_static_route_list),
                         ovn_api.add_static_route.call_count)
        del_route_calls = [mock.call(mock.ANY, ip_prefix=route['destination'],
                                     nexthop=route['nexthop'])
                           for route in del_static_route_list]
        ovn_api.delete_static_route.assert_has_calls(del_route_calls,
                                                     any_order=True)
        self.assertEqual(len(del_static_route_list),
                         ovn_api.delete_static_route.call_count)

        add_nat_calls = [mock.call(mock.ANY, **nat) for nat in add_snat_list]
        ovn_api.add_nat_rule_in_lrouter.assert_has_calls(add_nat_calls,
                                                         any_order=True)
        self.assertEqual(len(add_snat_list),
                         ovn_api.add_nat_rule_in_lrouter.call_count)

        add_fip_calls = [mock.call(nat, txn=mock.ANY)
                         for nat in add_floating_ip_list]
        (ovn_nb_synchronizer._ovn_client._create_or_update_floatingip.
            assert_has_calls(add_fip_calls))
        self.assertEqual(
            len(add_floating_ip_list),
            ovn_nb_synchronizer._ovn_client._create_or_update_floatingip.
            call_count)

        del_nat_calls = [mock.call(mock.ANY, **nat) for nat in del_snat_list]
        ovn_api.delete_nat_rule_in_lrouter.assert_has_calls(del_nat_calls,
                                                            any_order=True)
        self.assertEqual(len(del_snat_list),
                         ovn_api.delete_nat_rule_in_lrouter.call_count)

        del_fip_calls = [mock.call(nat, mock.ANY, txn=mock.ANY) for nat in
                         del_floating_ip_list]
        ovn_nb_synchronizer._ovn_client._delete_floatingip.assert_has_calls(
            del_fip_calls, any_order=True)
        self.assertEqual(
            len(del_floating_ip_list),
            ovn_nb_synchronizer._ovn_client._delete_floatingip.call_count)

        create_router_calls = [mock.call(mock.ANY, r,
                                         add_external_gateway=False)
                               for r in create_router_list]
        self.assertEqual(
            len(create_router_list),
            ovn_nb_synchronizer._ovn_client.create_router.call_count)
        ovn_nb_synchronizer._ovn_client.create_router.assert_has_calls(
            create_router_calls, any_order=True)

        create_router_port_calls = [
            mock.call(mock.ANY, self.routers[i], mock.ANY)
            for i, p in enumerate(create_router_port_list)]
        self.assertEqual(
            len(create_router_port_list),
            ovn_nb_synchronizer._ovn_client._create_lrouter_port.call_count)
        ovn_nb_synchronizer._ovn_client._create_lrouter_port.assert_has_calls(
            create_router_port_calls,
            any_order=True)

        self.assertEqual(len(del_router_list),
                         ovn_api.delete_lrouter.call_count)
        update_router_port_calls = [mock.call(mock.ANY, p)
                                    for p in update_router_port_list]
        self.assertEqual(
            len(update_router_port_list),
            ovn_nb_synchronizer._ovn_client.update_router_port.call_count)
        ovn_nb_synchronizer._ovn_client.update_router_port.assert_has_calls(
            update_router_port_calls,
            any_order=True)

        delete_lrouter_calls = [mock.call(r['router'])
                                for r in del_router_list]
        ovn_api.delete_lrouter.assert_has_calls(
            delete_lrouter_calls, any_order=True)

        self.assertEqual(
            len(del_router_port_list),
            ovn_api.delete_lrouter_port.call_count)
        delete_lrouter_port_calls = [mock.call(port['id'],
                                               port['router'], if_exists=False)
                                     for port in del_router_port_list]
        ovn_api.delete_lrouter_port.assert_has_calls(
            delete_lrouter_port_calls, any_order=True)

        self.assertEqual(
            len(add_subnet_dhcp_options_list),
            ovn_nb_synchronizer._ovn_client._add_subnet_dhcp_options.
            call_count)
        add_subnet_dhcp_options_calls = [
            mock.call(subnet, net, mock.ANY)
            for (subnet, net) in add_subnet_dhcp_options_list]
        ovn_nb_synchronizer._ovn_client._add_subnet_dhcp_options. \
            assert_has_calls(add_subnet_dhcp_options_calls, any_order=True)

        self.assertEqual(ovn_api.delete_dhcp_options.call_count,
                         len(delete_dhcp_options_list))
        delete_dhcp_options_calls = [
            mock.call(dhcp_opt_uuid)
            for dhcp_opt_uuid in delete_dhcp_options_list]
        ovn_api.delete_dhcp_options.assert_has_calls(
            delete_dhcp_options_calls, any_order=True)

    def test_ovn_nb_sync_mode_repair(self):

        create_network_list = [{'net': {'id': 'n2', 'mtu': 1450},
                                'ext_ids': {}}]
        del_network_list = ['neutron-n3']
        del_port_list = [{'id': 'p3n1', 'lswitch': 'neutron-n1'},
                         {'id': 'p1n1', 'lswitch': 'neutron-n1'},
                         {'id': 'provnet-orphaned-segment',
                          'lswitch': 'neutron-n4'}]
        create_port_list = self.ports
        for port in create_port_list.copy():
            if port['id'] in ['p1n1', 'fp1']:
                # this will be skipped by the logic,
                # because p1n1 is already in lswitch-port list
                # and fp1 is a floating IP port
                create_port_list.remove(port)
        create_provnet_port_list = [{'id': 'n1', 'mtu': 1450,
                                     'provider:physical_network': 'physnet1',
                                     'provider:segmentation_id': 1000}]
        create_router_list = [{
            'id': 'r2', 'routes': [
                {'nexthop': '40.0.0.100', 'destination': '30.0.0.0/24',
                 'external_ids': {}}],
            'gw_port_id': 'gpr2',
            'external_gateway_info': {
                'network_id': "ext-net", 'enable_snat': True,
                'external_fixed_ips': [{
                    'subnet_id': 'ext-subnet',
                    'ip_address': '100.0.0.2'}]}}]

        # Test adding and deleting routes snats fips behaviors for router r1
        # existing in both neutron DB and OVN DB.
        # Test adding behaviors for router r2 only existing in neutron DB.
        # Static routes with destination 0.0.0.0/0 are default gateway routes
        add_static_route_list = [{'nexthop': '20.0.0.101',
                                  'destination': '12.0.0.0/24',
                                  'external_ids': {}},
                                 {'nexthop': '90.0.0.1',
                                  'destination': '0.0.0.0/0',
                                  'external_ids': {
                                      ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                                      ovn_const.OVN_SUBNET_EXT_ID_KEY:
                                      'ext-subnet'}},
                                 {'nexthop': '40.0.0.100',
                                  'destination': '30.0.0.0/24',
                                  'external_ids': {}},
                                 {'nexthop': '100.0.0.1',
                                  'destination': '0.0.0.0/0',
                                  'external_ids': {
                                      ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                                      ovn_const.OVN_SUBNET_EXT_ID_KEY:
                                      'ext-subnet'}}]
        del_static_route_list = [{'nexthop': '20.0.0.100',
                                  'destination': '10.0.0.0/24',
                                  'external_ids': {}}]
        add_snat_list = [{'logical_ip': '172.16.2.0/24',
                          'external_ip': '90.0.0.2',
                          'type': 'snat'},
                         {'logical_ip': '192.168.2.0/24',
                          'external_ip': '100.0.0.2',
                          'type': 'snat'}]
        del_snat_list = [{'logical_ip': '172.16.1.0/24',
                          'external_ip': '90.0.0.2',
                          'type': 'snat'}]
        # fip 100.0.0.11 exists in OVN with distributed type and in Neutron
        # with centralized type. This fip is used to test
        # enable_distributed_floating_ip switch and migration
        add_floating_ip_list = [{'id': 'fip2', 'router_id': 'r1',
                                 'floating_ip_address': '90.0.0.12',
                                 'fixed_ip_address': '172.16.2.12'},
                                {'id': 'fip3', 'router_id': 'r2',
                                 'floating_ip_address': '100.0.0.10',
                                 'fixed_ip_address': '192.168.2.10'},
                                {'id': 'fip4', 'router_id': 'r2',
                                 'floating_ip_address': '100.0.0.11',
                                 'fixed_ip_address': '192.168.2.11'}]
        del_floating_ip_list = [{'logical_ip': '172.16.1.11',
                                 'external_ip': '90.0.0.11',
                                 'type': 'dnat_and_snat'},
                                {'logical_ip': '192.168.2.11',
                                 'external_ip': '100.0.0.11',
                                 'type': 'dnat_and_snat',
                                 'external_mac': '01:02:03:04:05:06',
                                 'logical_port': 'vm1'}]

        del_router_list = [{'router': 'neutron-r3'}]
        del_router_port_list = [{'id': 'lrp-p3r1', 'router': 'neutron-r1'}]
        create_router_port_list = self.get_sync_router_ports[:2]
        update_router_port_list = [self.get_sync_router_ports[2]]
        update_router_port_list[0].update(
            {'networks': self.lrport_networks})

        add_port_groups_list = [
            {'external_ids': {ovn_const.OVN_SG_EXT_ID_KEY: 'sg2'},
             'name': 'pg_sg2',
             'acls': []}]
        del_port_groups_list = ['pg_unknown_del']

        add_subnet_dhcp_options_list = [(self.subnets[2], self.networks[1]),
                                        (self.subnets[1], self.networks[0])]
        delete_dhcp_options_list = ['UUID2', 'UUID4', 'UUID5']

        ovn_nb_synchronizer = ovn_db_sync.OvnNbSynchronizer(
            self.plugin, self.mech_driver.nb_ovn, self.mech_driver.sb_ovn,
            'repair', self.mech_driver)
        self._test_ovn_nb_sync_helper(ovn_nb_synchronizer,
                                      self.networks,
                                      self.ports,
                                      self.routers,
                                      self.get_sync_router_ports,
                                      create_router_list,
                                      create_router_port_list,
                                      update_router_port_list,
                                      del_router_list, del_router_port_list,
                                      create_network_list, create_port_list,
                                      create_provnet_port_list,
                                      del_network_list, del_port_list,
                                      add_static_route_list,
                                      del_static_route_list,
                                      add_snat_list,
                                      del_snat_list,
                                      add_floating_ip_list,
                                      del_floating_ip_list,
                                      add_subnet_dhcp_options_list,
                                      delete_dhcp_options_list,
                                      add_port_groups_list,
                                      del_port_groups_list)

    def test_ovn_nb_sync_mode_log(self):
        create_network_list = []
        create_port_list = []
        create_provnet_port_list = []
        del_network_list = []
        del_port_list = []
        create_router_list = []
        create_router_port_list = []
        update_router_port_list = []
        del_router_list = []
        del_router_port_list = []
        add_static_route_list = []
        del_static_route_list = []
        add_snat_list = []
        del_snat_list = []
        add_floating_ip_list = []
        del_floating_ip_list = []
        add_subnet_dhcp_options_list = []
        delete_dhcp_options_list = []
        add_port_groups_list = []
        del_port_groups_list = []

        ovn_nb_synchronizer = ovn_db_sync.OvnNbSynchronizer(
            self.plugin, self.mech_driver.nb_ovn, self.mech_driver.sb_ovn,
            'log', self.mech_driver)
        self._test_ovn_nb_sync_helper(ovn_nb_synchronizer,
                                      self.networks,
                                      self.ports,
                                      self.routers,
                                      self.get_sync_router_ports,
                                      create_router_list,
                                      create_router_port_list,
                                      update_router_port_list,
                                      del_router_list, del_router_port_list,
                                      create_network_list, create_port_list,
                                      create_provnet_port_list,
                                      del_network_list, del_port_list,
                                      add_static_route_list,
                                      del_static_route_list,
                                      add_snat_list,
                                      del_snat_list,
                                      add_floating_ip_list,
                                      del_floating_ip_list,
                                      add_subnet_dhcp_options_list,
                                      delete_dhcp_options_list,
                                      add_port_groups_list,
                                      del_port_groups_list)

    def _test_ovn_nb_sync_calculate_routes_helper(self,
                                                  ovn_routes,
                                                  db_routes,
                                                  expected_added,
                                                  expected_deleted):
        ovn_nb_synchronizer = ovn_db_sync.OvnNbSynchronizer(
            self.plugin, self.mech_driver.nb_ovn, self.mech_driver.sb_ovn,
            'repair', self.mech_driver)
        add_routes, del_routes = ovn_nb_synchronizer. \
            _calculate_routes_differences(ovn_routes, db_routes)
        self.assertEqual(add_routes, expected_added)
        self.assertEqual(del_routes, expected_deleted)

    def test_ovn_nb_sync_calculate_routes_add_two_routes(self):

        # add 2 routes to ovn
        ovn_routes = []
        db_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}}]
        expected_added = db_routes
        expected_deleted = []
        self._test_ovn_nb_sync_calculate_routes_helper(ovn_routes,
                                                       db_routes,
                                                       expected_added,
                                                       expected_deleted)

    def test_ovn_nb_sync_calculate_routes_remove_two_routes(self):

        # remove 2 routes from ovn
        ovn_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}}]
        db_routes = []
        expected_added = []
        expected_deleted = ovn_routes
        self._test_ovn_nb_sync_calculate_routes_helper(ovn_routes,
                                                       db_routes,
                                                       expected_added,
                                                       expected_deleted)

    def test_ovn_nb_sync_calculate_routes_remove_and_add_two_routes(self):

        # remove 2 routes from ovn, add 2 routes to ovn
        ovn_routes = [{'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}},
                    {'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}}]
        db_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}}]
        expected_added = db_routes
        expected_deleted = ovn_routes
        self._test_ovn_nb_sync_calculate_routes_helper(ovn_routes,
                                                       db_routes,
                                                       expected_added,
                                                       expected_deleted)

    def test_ovn_nb_sync_calculate_routes_remove_and_keep_two_routes(self):

        # remove 2 routes from ovn, keep 2 routes
        ovn_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}},
                    {'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}}]
        db_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}}]
        expected_added = []
        expected_deleted = [{'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}},
                    {'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}}]
        self._test_ovn_nb_sync_calculate_routes_helper(ovn_routes,
                                                       db_routes,
                                                       expected_added,
                                                       expected_deleted)

    def test_ovn_nb_sync_calculate_routes_add_and_keep_two_routes(self):

        # add 2 routes to ovn, keep 2 routes
        ovn_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}}]
        db_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}},
                    {'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}}]
        expected_added = [{'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}},
                    {'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}}]
        expected_deleted = []
        self._test_ovn_nb_sync_calculate_routes_helper(ovn_routes,
                                                       db_routes,
                                                       expected_added,
                                                       expected_deleted)

    def test_ovn_nb_sync_calculate_routes_add_remove_keep_two_routes(self):

        # add 2 routes to ovn, remove 2 routes from ovn, keep 2 routes
        ovn_routes = [{'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}},
                    {'nexthop': '20.0.0.100',
                    'destination': '14.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '15.0.0.0/24',
                    'external_ids': {}}]
        db_routes = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '13.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '90.0.0.1',
                    'destination': '0.0.0.0/0',
                    'external_ids': {
                        ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
                        ovn_const.OVN_SUBNET_EXT_ID_KEY:
                        'ext-subnet'}}]

        expected_added = [{'nexthop': '20.0.0.100',
                    'destination': '11.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '12.0.0.0/24',
                    'external_ids': {}}]
        expected_deleted = [{'nexthop': '20.0.0.100',
                    'destination': '14.0.0.0/24',
                    'external_ids': {}},
                    {'nexthop': '20.0.0.100',
                    'destination': '15.0.0.0/24',
                    'external_ids': {}}]
        self._test_ovn_nb_sync_calculate_routes_helper(ovn_routes,
                                                       db_routes,
                                                       expected_added,
                                                       expected_deleted)


class TestIsRouterPortChanged(test_mech_driver.OVNMechanismDriverTestCase):

    def setUp(self):
        super(TestIsRouterPortChanged, self).setUp()
        self.ovn_nb_synchronizer = ovn_db_sync.OvnNbSynchronizer(
            self.plugin, self.mech_driver.nb_ovn, self.mech_driver.sb_ovn,
            'log', self.mech_driver)

        self.db_router_port = {
            'id': 'aa076509-915d-4b1c-8d9d-3db53d9c5faf',
            'networks': ['fdf9:ad62:3a04::1/64'],
            'ipv6_ra_configs': {'address_mode': 'slaac',
                                'send_periodic': 'true',
                                'mtu': '1442'}
        }
        self.lrport_nets = ['fdf9:ad62:3a04::1/64']
        self.ovn_lrport = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'ipv6_ra_configs': {'address_mode': 'slaac',
                                       'send_periodic': 'true',
                                       'mtu': '1442'}})

        self.ovn_nb_synchronizer.ovn_api.is_col_present.return_value = True
        self.ovn_nb_synchronizer.ovn_api.lrp_get().execute.return_value = (
            self.ovn_lrport)

    def test__is_router_port_changed_not_changed(self):
        self.assertFalse(self.ovn_nb_synchronizer._is_router_port_changed(
            self.db_router_port, self.lrport_nets))

    def test__is_router_port_changed_network_changed(self):
        self.db_router_port['networks'] = ['172.24.4.26/24',
                                           '2001:db8::206/64']
        self.assertTrue(self.ovn_nb_synchronizer._is_router_port_changed(
            self.db_router_port, self.lrport_nets))

    def test__is_router_port_changed_ipv6_ra_configs_changed(self):
        self.db_router_port['ipv6_ra_configs']['mtu'] = '1500'
        self.assertTrue(self.ovn_nb_synchronizer._is_router_port_changed(
            self.db_router_port, self.lrport_nets))


class TestOvnSbSyncML2(test_mech_driver.OVNMechanismDriverTestCase):

    def test_ovn_sb_sync(self):
        ovn_sb_synchronizer = ovn_db_sync.OvnSbSynchronizer(
            self.plugin,
            self.mech_driver.sb_ovn,
            self.mech_driver)
        ovn_api = ovn_sb_synchronizer.ovn_api
        hostname_with_physnets = {'hostname1': ['physnet1', 'physnet2'],
                                  'hostname2': ['physnet1']}
        ovn_api.get_chassis_hostname_and_physnets.return_value = (
            hostname_with_physnets)
        ovn_driver = ovn_sb_synchronizer.ovn_driver
        ovn_driver.update_segment_host_mapping = mock.Mock()
        hosts_in_neutron = {'hostname2', 'hostname3'}

        with mock.patch.object(ovn_db_sync.segments_db,
                               'get_hosts_mapped_with_segments',
                               return_value=hosts_in_neutron) as mock_ghmws:
            ovn_sb_synchronizer.sync_hostname_and_physical_networks(mock.ANY)
            mock_ghmws.assert_called_once_with(
                mock.ANY, include_agent_types={ovn_const.OVN_CONTROLLER_AGENT})
            all_hosts = set(hostname_with_physnets.keys()) | hosts_in_neutron
            self.assertEqual(
                len(all_hosts),
                ovn_driver.update_segment_host_mapping.call_count)
            update_segment_host_mapping_calls = [mock.call(
                host, hostname_with_physnets[host])
                for host in hostname_with_physnets]
            update_segment_host_mapping_calls += [
                mock.call(host, []) for host in
                hosts_in_neutron - set(hostname_with_physnets.keys())]
            ovn_driver.update_segment_host_mapping.assert_has_calls(
                update_segment_host_mapping_calls, any_order=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import pprint
import time

from neutron_lib.plugins import directory
from oslo_log import log as logging
from oslo_utils import importutils

from neutron.api.rpc.callbacks import exceptions

LOG = logging.getLogger(__name__)

VERSIONS_TTL = 60


# NOTE(mangelajo): if we import this globally we end up with a (very
#                  long) circular dependency, this can be fixed if we
#                  stop importing all exposed classes in
#                  neutron.api.rpc.callbacks.resources and provide
#                  a decorator to expose classes
def _import_resources():
    return importutils.import_module('neutron.api.rpc.callbacks.resources')


def _import_agents_db():
    return importutils.import_module('neutron.db.agents_db')


AgentConsumer = collections.namedtuple('AgentConsumer', ['agent_type',
                                                         'host'])
AgentConsumer.__repr__ = lambda self: '%s@%s' % self


class ResourceConsumerTracker(object):
    """Class passed down to collect consumer's resource versions.

    This class is responsible for fetching the local versions of
    resources, and letting the called function register every consumer's
    resource version.

    This class is passed down to the plugin get_agents_resource_versions
    currently, as the only expected consumers are agents so far.

    Later on, this class can also be used to recalculate, for each
    resource type, the collection of versions that are local or
    known by one or more consumers.
    """

    def __init__(self):
        # Initialize with the local (server) versions, as we always want
        # to send those. Agents, as they upgrade, will need the latest version,
        # and there is a corner case we'd not be covering otherwise:
        #   1) one or several neutron-servers get disconnected from rpc (while
        #      running)
        #   2) a new agent comes up, with the latest version and it reports
        #      2 ways:
        #     a) via status report (which will be stored in the database)
        #     b) via fanout call to all neutron servers, this way, all of them
        #        get their version set updated right away without the need to
        #        re-fetch anything from the database.
        #   3) the neutron-servers get back online to the rpc bus, but they
        #      lost the fanout message.
        #
        # TODO(mangelajo) To cover this case we may need a callback from oslo
        # messaging to get notified about disconnections/reconnections to the
        # rpc bus, invalidating the consumer version cache when we receive such
        # callback.
        self._versions = self._get_local_resource_versions()
        self._versions_by_consumer = collections.defaultdict(dict)
        self._needs_recalculation = False
        self.last_report = None

    def _get_local_resource_versions(self):
        resources = _import_resources()
        local_resource_versions = collections.defaultdict(set)
        for resource_type, version in (
                resources.LOCAL_RESOURCE_VERSIONS.items()):
            local_resource_versions[resource_type].add(version)
        return local_resource_versions

    # TODO(mangelajo): add locking with _recalculate_versions if we ever
    #                  move out of green threads.
    def _set_version(self, consumer, resource_type, version):
        """Set or update a consumer resource type version."""
        self._versions[resource_type].add(version)
        consumer_versions = self._versions_by_consumer[consumer]
        prev_version = consumer_versions.get(resource_type, None)
        if version:
            consumer_versions[resource_type] = version
        else:
            consumer_versions.pop(resource_type, None)

        if prev_version != version:
            # If a version got updated/changed in a consumer, we need to
            # recalculate the main dictionary of versions based on the
            # new _versions_by_consumer.
            # We defer the recalculation until every consumer version has
            # been set for all of its resource types.
            self._needs_recalculation = True
            LOG.debug("Version for resource type %(resource_type)s changed "
                      "%(prev_version)s to %(version)s on "
                      "consumer %(consumer)s",
                      {'resource_type': resource_type,
                       'version': version,
                       'prev_version': prev_version,
                       'consumer': consumer})

    def set_versions(self, consumer, versions):
        """Set or update an specific consumer resource types.

        :param consumer: should be an AgentConsumer object, with agent_type
                         and host set. This acts as the unique ID for the
                         agent.
        :param versions: should be a dictionary in the following format:
                               {'QosPolicy': '1.1',
                                'SecurityGroup': '1.0',
                                'Port': '1.0'}
        """
        for resource_type, resource_version in versions.items():
            self._set_version(consumer, resource_type,
                              resource_version)

        if versions:
            self._cleanup_removed_versions(consumer, versions)
        else:
            self._handle_no_set_versions(consumer)

    def _cleanup_removed_versions(self, consumer, versions):
        """Check if any version report has been removed, and cleanup."""
        prev_resource_types = set(
            self._versions_by_consumer[consumer].keys())
        cur_resource_types = set(versions.keys())
        removed_resource_types = prev_resource_types - cur_resource_types
        if removed_resource_types:
            LOG.debug("Removing stale tracked versions: %s",
                      removed_resource_types)
        for resource_type in removed_resource_types:
            self._set_version(consumer, resource_type, None)

    def _handle_no_set_versions(self, consumer):
        """Handle consumers reporting no versions."""
        if self._versions_by_consumer[consumer]:
            self._needs_recalculation = True
            LOG.debug("Clearing versions for consumer %s", consumer)
        self._versions_by_consumer[consumer] = {}

    def get_resource_versions(self, resource_type):
        """Fetch the versions necessary to notify all consumers."""
        if self._needs_recalculation:
            self._recalculate_versions()
            self._needs_recalculation = False

        return copy.copy(self._versions[resource_type])

    def report(self):
        """Output debug information about the consumer versions."""

        def _format(versions):
            return pprint.pformat(dict(versions), indent=4)

        debug_dict = {'pushed_versions': _format(self._versions),
                      'consumer_versions': _format(self._versions_by_consumer)}
        if self.last_report != debug_dict:
            self.last_report = debug_dict
            LOG.debug('Tracked resource versions report:\n'
                      'pushed versions:\n%(pushed_versions)s\n\n'
                      'consumer versions:\n%(consumer_versions)s\n',
                      debug_dict)

    # TODO(mangelajo): Add locking if we ever move out of greenthreads.
    def _recalculate_versions(self):
        """Recalculate the _versions set.

        Re-fetch the local (server) versions and expand with consumers'
        versions.
        """
        versions = self._get_local_resource_versions()
        for versions_dict in self._versions_by_consumer.values():
            for res_type, res_version in versions_dict.items():
                versions[res_type].add(res_version)
        self._versions = versions


class CachedResourceConsumerTracker(object):
    """This class takes care of the caching logic of versions."""

    def __init__(self):
        # This is TTL expiration time, 0 means it will be expired at start
        self._expires_at = 0
        self._versions = ResourceConsumerTracker()

    def _update_consumer_versions(self):
        new_tracker = ResourceConsumerTracker()
        neutron_plugin = directory.get_plugin()
        agents_db = _import_agents_db()
        # If you use RPC callbacks, your plugin needs to implement
        # AgentsDbMixin so that we know which resource versions your
        # agents consume via RPC, please note that rpc_callbacks are
        # only designed to work with agents currently.
        if isinstance(neutron_plugin, agents_db.AgentDbMixin):
            neutron_plugin.get_agents_resource_versions(new_tracker)
        else:
            raise exceptions.NoAgentDbMixinImplemented()

        # preserve last report state so we don't duplicate logs on refresh
        new_tracker.last_report = self._versions.last_report
        self._versions = new_tracker
        self._versions.report()

    def _check_expiration(self):
        if time.time() > self._expires_at:
            self._update_consumer_versions()
            self._expires_at = time.time() + VERSIONS_TTL

    def get_resource_versions(self, resource_type):
        self._check_expiration()
        return self._versions.get_resource_versions(resource_type)

    def update_versions(self, consumer, resource_versions):
        self._versions.set_versions(consumer, resource_versions)

    def report(self):
        self._check_expiration()
        self._versions.report()


_cached_version_tracker = None


# NOTE(ajo): add locking if we ever stop using greenthreads
def _get_cached_tracker():
    global _cached_version_tracker
    if not _cached_version_tracker:
        _cached_version_tracker = CachedResourceConsumerTracker()
    return _cached_version_tracker


def get_resource_versions(resource_type):
    """Return the set of versions expected by the consumers of a resource."""
    return _get_cached_tracker().get_resource_versions(resource_type)


def update_versions(consumer, resource_versions):
    """Update the resources' versions for a consumer id."""
    _get_cached_tracker().update_versions(consumer, resource_versions)


def report():
    """Report resource versions in debug logs."""
    _get_cached_tracker().report()



# Copyright (C) 2013 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.agent import topics
from neutron_lib.api import extensions
from neutron_lib import constants
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_log import log as logging
import oslo_messaging

from neutron.db import agentschedulers_db

LOG = logging.getLogger(__name__)


class MeteringAgentNotifyAPI(object):
    """API for plugin to notify L3 metering agent."""

    def __init__(self, topic=topics.METERING_AGENT):
        self.topic = topic
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def _agent_notification(self, context, method, routers):
        """Notify l3 metering agents hosted by l3 agent hosts."""
        adminContext = context if context.is_admin else context.elevated()
        plugin = directory.get_plugin(plugin_constants.L3)

        l3_routers = {}
        state = agentschedulers_db.get_admin_state_up_filter()
        for router in routers:
            l3_agents = plugin.get_l3_agents_hosting_routers(
                adminContext, [router['id']],
                admin_state_up=state,
                active=True)
            for l3_agent in l3_agents:
                LOG.debug('Notify metering agent at %(topic)s.%(host)s '
                          'the message %(method)s',
                          {'topic': self.topic,
                           'host': l3_agent.host,
                           'method': method})

                l3_router = l3_routers.get(l3_agent.host, [])
                l3_router.append(router)
                l3_routers[l3_agent.host] = l3_router

        for host, routers in l3_routers.items():
            cctxt = self.client.prepare(server=host)
            cctxt.cast(context, method, routers=routers)

    def _notification_fanout(self, context, method, router_id):
        LOG.debug('Fanout notify metering agent at %(topic)s the message '
                  '%(method)s on router %(router_id)s',
                  {'topic': self.topic,
                   'method': method,
                   'router_id': router_id})
        cctxt = self.client.prepare(fanout=True)
        cctxt.cast(context, method, router_id=router_id)

    def _notification_host(self, context, method, host, **kwargs):
        """Notify the agent that is hosting the router."""
        LOG.debug('Notify agent at %(host)s the message '
                  '%(method)s', {'host': host,
                                 'method': method})
        cctxt = self.client.prepare(server=host)
        cctxt.cast(context, method, **kwargs)

    def _notification(self, context, method, routers):
        """Notify all the agents that are hosting the routers."""
        plugin = directory.get_plugin(plugin_constants.L3)
        if extensions.is_extension_supported(
                plugin, constants.L3_AGENT_SCHEDULER_EXT_ALIAS):
            self._agent_notification(context, method, routers)
        else:
            cctxt = self.client.prepare(fanout=True)
            cctxt.cast(context, method, routers=routers)

    def router_deleted(self, context, router_id):
        self._notification_fanout(context, 'router_deleted', router_id)

    def routers_updated(self, context, routers):
        if routers:
            self._notification(context, 'routers_updated', routers)

    def update_metering_label_rules(self, context, routers):
        self._notification(context, 'update_metering_label_rules', routers)

    def add_metering_label_rule(self, context, routers):
        self._notification(context, 'add_metering_label_rule', routers)

    def remove_metering_label_rule(self, context, routers):
        self._notification(context, 'remove_metering_label_rule', routers)

    def add_metering_label(self, context, routers):
        self._notification(context, 'add_metering_label', routers)

    def remove_metering_label(self, context, routers):
        self._notification(context, 'remove_metering_label', routers)

    def routers_updated_on_host(self, context, router_ids, host):
        """Notify router updates to specific hosts hosting DVR routers."""
        self._notification_host(context, 'routers_updated', host,
                                routers=router_ids)



# Copyright 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil
import tempfile

from neutron_lib import constants

from neutron.common import utils
from neutron.plugins.ml2.extensions import qos as qos_ext
from neutron.tests import base
from neutron.tests.common import config_fixtures
from neutron.tests.common.exclusive_resources import port
from neutron.tests.common import helpers as c_helpers
from neutron.tests.common import net_helpers
from neutron.tests.fullstack import base as fullstack_base

PHYSICAL_NETWORK_NAME = "physnet1"
MINIMUM_BANDWIDTH_INGRESS_KBPS = 1000
MINIMUM_BANDWIDTH_EGRESS_KBPS = 1000

NEUTRON_SERVER_PORT_START = 10000
NEUTRON_SERVER_PORT_END = 20000

OVS_OF_PORT_LISTEN_START = 20001
OVS_OF_PORT_LISTEN_END = 30000

CLIENT_CONN_PORT_START = 30001
CLIENT_CONN_PORT_END = 65000


class ConfigFixture(config_fixtures.ConfigFileFixture):
    """A fixture that holds an actual Neutron configuration.

    Note that 'self.config' is intended to only be updated once, during
    the constructor, so if this fixture is re-used (setUp is called twice),
    then the dynamic configuration values won't change. The correct usage
    is initializing a new instance of the class.
    """
    def __init__(self, env_desc, host_desc, temp_dir, base_filename):
        super(ConfigFixture, self).__init__(
            base_filename, config_fixtures.ConfigDict(), temp_dir)
        self.env_desc = env_desc
        self.host_desc = host_desc

    def _generate_namespace_suffix(self):
        return utils.get_rand_name(prefix='test')


class NeutronConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir,
                 connection, rabbitmq_environment, use_local_apipaste=True):
        super(NeutronConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, base_filename='neutron.conf')

        self.config.update({
            'DEFAULT': {
                'host': self._generate_host(),
                'state_path': self._generate_state_path(self.temp_dir),
                'core_plugin': 'ml2',
                'service_plugins': env_desc.service_plugins,
                'auth_strategy': 'noauth',
                'debug': 'True',
                'global_physnet_mtu': str(env_desc.global_mtu),
                'agent_down_time': str(env_desc.agent_down_time),
                'transport_url':
                    'rabbit://%(user)s:%(password)s@%(host)s:5672/%(vhost)s' %
                    {'user': rabbitmq_environment.user,
                     'password': rabbitmq_environment.password,
                     'host': rabbitmq_environment.host,
                     'vhost': rabbitmq_environment.vhost},
                'api_workers': str(env_desc.api_workers),
            },
            'database': {
                'connection': connection,
            },
            'oslo_concurrency': {
                'lock_path': '$state_path/lock',
            },
            'agent': {
                'report_interval': str(env_desc.agent_down_time // 2),
                'log_agent_heartbeats': 'True',
            },
            'quotas': {
                'quota_driver': env_desc.quota_driver
            },
            'experimental': {
                'linuxbridge': str(env_desc.allow_experimental_linuxbridge)
            },
        })

        if use_local_apipaste:
            self.config['DEFAULT']['api_paste_config'] = (
                self._generate_api_paste())

        policy_file = self._generate_policy_yaml()
        if policy_file:
            self.config['oslo_policy'] = {'policy_file': policy_file}

        # Set root_helper/root_helper_daemon only when env var is set
        root_helper = os.environ.get('OS_ROOTWRAP_CMD')
        if root_helper:
            self.config['agent']['root_helper'] = root_helper
        root_helper_daemon = os.environ.get('OS_ROOTWRAP_DAEMON_CMD')
        if root_helper_daemon:
            self.config['agent']['root_helper_daemon'] = root_helper_daemon
        if env_desc.router_scheduler:
            self.config['DEFAULT']['router_scheduler_driver'] = (
                env_desc.router_scheduler)
        if env_desc.has_placement:
            service_plugins = self.config['DEFAULT']['service_plugins']
            self.config['DEFAULT']['service_plugins'] = (
                '%s,%s' % (service_plugins, 'placement')
            )
            self.config.update({
                'placement': {
                    'auth_type': 'noauth',
                    'auth_section': 'http://127.0.0.1:%s/placement' %
                    env_desc.placement_port
                }
            })
        if env_desc.dhcp_scheduler_class:
            self.config['DEFAULT']['dhcp_agents_per_network'] = '1'
            self.config['DEFAULT']['network_scheduler_driver'] = (
                env_desc.dhcp_scheduler_class)

        self.config['DEFAULT']['enable_traditional_dhcp'] = str(
            env_desc.enable_traditional_dhcp)

        net_helpers.set_local_port_range(CLIENT_CONN_PORT_START,
                                         CLIENT_CONN_PORT_END)

    def _setUp(self):
        self.config['DEFAULT'].update({
            'bind_port': self.useFixture(
                port.ExclusivePort(constants.PROTO_NAME_TCP,
                                   start=NEUTRON_SERVER_PORT_START,
                                   end=NEUTRON_SERVER_PORT_END)).port
        })
        super(NeutronConfigFixture, self)._setUp()

    def _generate_host(self):
        return utils.get_rand_name(prefix='host-')

    def _generate_state_path(self, temp_dir):
        # Assume that temp_dir will be removed by the caller
        self.state_path = tempfile.mkdtemp(prefix='state_path', dir=temp_dir)
        return self.state_path

    def _generate_api_paste(self):
        return c_helpers.find_sample_file('api-paste.ini')

    def _generate_policy_yaml(self):
        return c_helpers.find_sample_file('policy.yaml')

    def get_host(self):
        return self.config['DEFAULT']['host']


class ML2ConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, tenant_network_types):
        super(ML2ConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, base_filename='ml2_conf.ini')

        mechanism_drivers = self.env_desc.mech_drivers
        if self.env_desc.l2_pop:
            mechanism_drivers += ',l2population'

        net_vlan_ranges_extra = ''
        if 'segments' in env_desc.service_plugins:
            net_vlan_ranges_extra = (',' + PHYSICAL_NETWORK_NAME +
                                     '_lb:1050:1059')

        self.config.update({
            'ml2': {
                'tenant_network_types': tenant_network_types,
                'mechanism_drivers': mechanism_drivers,
            },
            'ml2_type_vlan': {
                'network_vlan_ranges': PHYSICAL_NETWORK_NAME + ':1000:1029' +
                net_vlan_ranges_extra,
            },
            'ml2_type_gre': {
                'tunnel_id_ranges': '1:30',
            },
            'ml2_type_vxlan': {
                'vni_ranges': '1001:1030',
            },
        })

        extension_drivers = {'port_security'}
        if env_desc.qos:
            extension_drivers.add(qos_ext.QOS_EXT_DRIVER_ALIAS)
        if env_desc.ml2_extension_drivers:
            extension_drivers.update(env_desc.ml2_extension_drivers)
        self.config['ml2']['extension_drivers'] = ','.join(extension_drivers)


class OVSConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, local_ip, **kwargs):
        super(OVSConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir,
            base_filename='openvswitch_agent.ini')

        self.tunneling_enabled = self.env_desc.tunneling_enabled
        ext_dev = utils.get_rand_device_name(prefix='br-eth')
        self.config.update({
            'ovs': {
                'local_ip': local_ip,
                'integration_bridge': self._generate_integration_bridge(),
                'bridge_mappings': '%s:%s' % (PHYSICAL_NETWORK_NAME, ext_dev),
                'of_inactivity_probe': '0',
                'ovsdb_debug': 'True',
            },
            'securitygroup': {
                'firewall_driver': host_desc.firewall_driver,
            },
            'agent': {
                'l2_population': str(self.env_desc.l2_pop),
                'arp_responder': str(self.env_desc.arp_responder),
                'debug_iptables_rules': str(env_desc.debug_iptables),
                'use_helper_for_ns_read': 'False',
            }
        })

        if self.tunneling_enabled:
            self.config['agent'].update({
                'tunnel_types': self.env_desc.network_type})
            self.config['ovs'].update({
                'tunnel_bridge': self._generate_tunnel_bridge(),
                'int_peer_patch_port': self._generate_int_peer(),
                'tun_peer_patch_port': self._generate_tun_peer()})
        else:
            if env_desc.report_bandwidths:
                self.config['ovs'][constants.RP_BANDWIDTHS] = \
                    '%s:%s:%s' % (ext_dev, MINIMUM_BANDWIDTH_EGRESS_KBPS,
                                  MINIMUM_BANDWIDTH_INGRESS_KBPS)

        if env_desc.qos:
            self.config['agent']['extensions'] = 'qos'
        if env_desc.log:
            self.config['agent']['extensions'] = 'log'
            test_name = kwargs.get("test_name")
            test_name = base.sanitize_log_path(test_name)
            self.config.update({
                'network_log': {
                    'local_output_log_base':
                        self._generate_temp_log_file(test_name)}
            })
        if not env_desc.enable_traditional_dhcp:
            self.config['agent']['extensions'] = 'dhcp'
            self.config.update({
                'dhcp': {
                    'enable_ipv6': 'True',
                    'dhcp_renewal_time': '0',
                    'dhcp_rebinding_time': '0'}
            })
        if env_desc.local_ip_ext:
            self.config['agent']['extensions'] = 'local_ip'
            if host_desc.firewall_driver == 'openvswitch':
                self.config['local_ip'] = {'static_nat': 'True'}

    def _setUp(self):
        self.config['ovs'].update({
            'of_listen_port': self.useFixture(
                port.ExclusivePort(constants.PROTO_NAME_TCP,
                                   start=OVS_OF_PORT_LISTEN_START,
                                   end=OVS_OF_PORT_LISTEN_END)).port
        })
        super(OVSConfigFixture, self)._setUp()

    def _generate_integration_bridge(self):
        return utils.get_rand_device_name(prefix='br-int')

    def _generate_tunnel_bridge(self):
        return utils.get_rand_device_name(prefix='br-tun')

    def _generate_int_peer(self):
        return utils.get_rand_device_name(prefix='patch-tun')

    def _generate_tun_peer(self):
        return utils.get_rand_device_name(prefix='patch-int')

    def _generate_temp_log_file(self, test_name):
        log_dir_path = os.path.join(fullstack_base.DEFAULT_LOG_DIR, test_name)
        if not os.path.exists(log_dir_path):
            os.mkdir(log_dir_path, 0o755)
        return '%s/%s.log' % (log_dir_path,
                              utils.get_rand_name(prefix="test-sg-"))

    def get_br_int_name(self):
        return self.config.ovs.integration_bridge

    def get_br_phys_name(self):
        return self.config.ovs.bridge_mappings.split(':')[1]

    def get_br_tun_name(self):
        return self.config.ovs.tunnel_bridge


class SRIOVConfigFixture(ConfigFixture):
    def __init__(self, env_desc, host_desc, temp_dir, local_ip):
        super(SRIOVConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir,
            base_filename='sriov_agent.ini')

        device1 = utils.get_rand_device_name(prefix='ens5')
        device2 = utils.get_rand_device_name(prefix='ens6')
        phys_dev_mapping = '%s:%s,%s:%s' % (PHYSICAL_NETWORK_NAME, device1,
                                            PHYSICAL_NETWORK_NAME, device2)
        rp_bandwidths = '%s:%s:%s,%s:%s:%s' % (device1,
                                               MINIMUM_BANDWIDTH_EGRESS_KBPS,
                                               MINIMUM_BANDWIDTH_INGRESS_KBPS,
                                               device2,
                                               MINIMUM_BANDWIDTH_EGRESS_KBPS,
                                               MINIMUM_BANDWIDTH_INGRESS_KBPS)
        self.config.update({
            'sriov_nic': {
                'physical_device_mappings': phys_dev_mapping,
                'resource_provider_bandwidths': rp_bandwidths,
            }
        })

    def _setUp(self):
        super(SRIOVConfigFixture, self)._setUp()


class PlacementConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir):
        super(PlacementConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, base_filename='placement.ini')
        self.config.update({
            'DEFAULT': {
                'debug': 'True',
                'placement_port': self.env_desc.placement_port
            }
        })

    def _setUp(self):
        super(PlacementConfigFixture, self)._setUp()


class LinuxBridgeConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, local_ip,
                 physical_device_name):
        super(LinuxBridgeConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir,
            base_filename="linuxbridge_agent.ini"
        )
        self.service_plugins = env_desc.service_plugins

        self.config.update({
            'VXLAN': {
                'enable_vxlan': str(self.env_desc.tunneling_enabled),
                'local_ip': local_ip,
                'l2_population': str(self.env_desc.l2_pop),
            },
            'securitygroup': {
                'firewall_driver': host_desc.firewall_driver,
            },
            'AGENT': {
                'debug_iptables_rules': str(env_desc.debug_iptables),
                'use_helper_for_ns_read': 'False',
            }
        })
        if env_desc.qos:
            self.config.update({
                'AGENT': {
                    'extensions': 'qos'
                }
            })
        if self.env_desc.tunneling_enabled:
            self.config.update({
                'LINUX_BRIDGE': {
                    'bridge_mappings': self._generate_bridge_mappings(
                        physical_device_name
                    )
                }
            })
        else:
            self.config.update({
                'LINUX_BRIDGE': {
                    'physical_interface_mappings':
                        self._generate_bridge_mappings(
                            physical_device_name
                        )
                }
            })

    def _generate_bridge_mappings(self, device_name):
        bridge_mappings_extra = ('_lb' if 'segments' in self.service_plugins
                                 else '')
        return '%s%s:%s' % (PHYSICAL_NETWORK_NAME, bridge_mappings_extra,
                            device_name)


class L3ConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, integration_bridge=None):
        super(L3ConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, base_filename='l3_agent.ini')
        if host_desc.l2_agent_type == constants.AGENT_TYPE_OVS:
            self._prepare_config_with_ovs_agent(integration_bridge)
        elif host_desc.l2_agent_type == constants.AGENT_TYPE_LINUXBRIDGE:
            self._prepare_config_with_linuxbridge_agent()
        if host_desc.l3_agent_mode:
            self.config['DEFAULT'].update({
                'agent_mode': host_desc.l3_agent_mode})
        self.config['DEFAULT'].update({
            'debug': 'True',
            'test_namespace_suffix': self._generate_namespace_suffix(),
        })
        self.config.update({
            'agent': {'use_helper_for_ns_read': 'False'}
        })
        if host_desc.availability_zone:
            self.config['agent'].update({
                'availability_zone': host_desc.availability_zone
            })
        if host_desc.l3_agent_extensions:
            self.config['agent'].update({
                'extensions': host_desc.l3_agent_extensions
            })

    def _prepare_config_with_ovs_agent(self, integration_bridge):
        self.config.update({
            'DEFAULT': {
                'interface_driver': ('neutron.agent.linux.interface.'
                                     'OVSInterfaceDriver'),
            },
            'OVS': {
                'integration_bridge': integration_bridge,
            }
        })

    def _prepare_config_with_linuxbridge_agent(self):
        self.config.update({
            'DEFAULT': {
                'interface_driver': ('neutron.agent.linux.interface.'
                                     'BridgeInterfaceDriver'),
            }
        })


class DhcpConfigFixture(ConfigFixture):

    def __init__(self, env_desc, host_desc, temp_dir, integration_bridge=None):
        super(DhcpConfigFixture, self).__init__(
            env_desc, host_desc, temp_dir, base_filename='dhcp_agent.ini')

        if host_desc.l2_agent_type == constants.AGENT_TYPE_OVS:
            self._prepare_config_with_ovs_agent(integration_bridge)
        elif host_desc.l2_agent_type == constants.AGENT_TYPE_LINUXBRIDGE:
            self._prepare_config_with_linuxbridge_agent()
        self.config['DEFAULT'].update({
            'debug': 'True',
            'dhcp_confs': self._generate_dhcp_path(),
            'test_namespace_suffix': self._generate_namespace_suffix()
        })
        self.config.update({
            'AGENT': {'use_helper_for_ns_read': 'False'}
        })
        if host_desc.availability_zone:
            self.config['AGENT'].update({
                'availability_zone': host_desc.availability_zone
            })

    def _setUp(self):
        super(DhcpConfigFixture, self)._setUp()
        self.addCleanup(self._clean_dhcp_path)

    def _prepare_config_with_ovs_agent(self, integration_bridge):
        self.config.update({
            'DEFAULT': {
                'interface_driver': 'openvswitch',
            },
            'OVS': {
                'integration_bridge': integration_bridge,
            }
        })

    def _prepare_config_with_linuxbridge_agent(self):
        self.config.update({
            'DEFAULT': {
                'interface_driver': 'linuxbridge',
            }
        })

    def _generate_dhcp_path(self):
        # NOTE(slaweq): dhcp_conf path needs to be directory with read
        # permission for everyone, otherwise dnsmasq process will not be able
        # to read his configs
        self.dhcp_path = tempfile.mkdtemp(prefix="dhcp_configs_", dir="/tmp/")
        os.chmod(self.dhcp_path, 0o755)
        return self.dhcp_path

    def _clean_dhcp_path(self):
        shutil.rmtree(self.dhcp_path, ignore_errors=True)



# Copyright 2015 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions

from neutron._i18n import _


class InvalidSubnetRequestType(exceptions.BadRequest):
    message = _("Cannot handle subnet of type %(subnet_type)s")


class AddressCalculationFailure(exceptions.NeutronException):
    message = _("Unable to calculate %(address_type)s address because of:"
                "%(reason)s")


class InvalidAddressType(exceptions.NeutronException):
    message = _("Unknown address type %(address_type)s")


class IpAddressAlreadyAllocated(exceptions.Conflict):
    message = _("IP address %(ip)s already allocated in subnet %(subnet_id)s")


class InvalidIpForSubnet(exceptions.BadRequest):
    message = _("IP address %(ip)s does not belong to subnet %(subnet_id)s")


class InvalidAddressRequest(exceptions.BadRequest):
    message = _("The address allocation request could not be satisfied "
                "because: %(reason)s")


class InvalidSubnetRequest(exceptions.BadRequest):
    message = _("The subnet request could not be satisfied because: "
                "%(reason)s")


class IPAddressChangeNotAllowed(exceptions.BadRequest):
    message = _("IP updates for port %(port_id)s are not allowed")


class AllocationOnAutoAddressSubnet(exceptions.InvalidInput):
    message = _("IPv6 address %(ip)s cannot be directly "
                "assigned to a port on subnet %(subnet_id)s as the "
                "subnet is configured for automatic addresses")


class IpAddressGenerationFailure(exceptions.Conflict):
    message = _("No more IP addresses available for subnet %(subnet_id)s.")


class IpAddressGenerationFailureAllSubnets(IpAddressGenerationFailure):
    message = _("No more IP addresses available.")


class IpAddressGenerationFailureNoMatchingSubnet(IpAddressGenerationFailure):
    message = _("No valid service subnet for the given device owner, "
                "network %(network_id)s, service type %(service_type)s.")


class IPAllocationFailed(exceptions.NeutronException):
    message = _("IP allocation failed. Try again later.")


class IpamValueInvalid(exceptions.Conflict):
    def __init__(self, message=None):
        self.message = message
        super(IpamValueInvalid, self).__init__()


class DeferIpam(exceptions.NeutronException):
    message = _("Exception used to signal that IP allocation is deferred")



# Copyright (c) 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from neutron.agent.ovn.agent import ovsdb as agent_ovsdb
from neutron.agent.ovn.extensions import qos_hwol
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.common import utils as n_utils
from neutron.tests import base as test_base
from neutron.tests.functional import base


class OVSInterfaceEventTestCase(base.TestOVNFunctionalBase):

    def _cleanup(self):
        self.ovs_idl.del_port(self.port_name, bridge=self.br_name).execute(
            check_error=False)
        self.ovs_idl.del_br(self.br_name).execute(check_error=False)

    @test_base.unstable_test('bug 2006603')
    def test_port_creation_and_deletion(self):
        def check_add_port_called():
            try:
                mock_agent[qos_hwol.EXT_NAME].add_port.assert_has_calls(
                    [mock.call(port_iface_id, self.port_name)])
                return True
            except AssertionError:
                return False

        def check_remove_egress_called():
            try:
                mock_agent.qos_hwol_ext.remove_egress.assert_has_calls(
                    [mock.call(port_iface_id)])
                return True
            except AssertionError:
                return False

        port_iface_id = 'port_iface-id'
        mock_agent = mock.MagicMock()
        events = [qos_hwol.OVSInterfaceEvent(mock_agent)]
        self.ovs_idl = agent_ovsdb.MonitorAgentOvsIdl(events=events).start()
        self.br_name = ('brtest-' + uuidutils.generate_uuid())[:13]
        self.port_name = ('port-' + uuidutils.generate_uuid())[:13]
        self.addCleanup(self._cleanup)
        with self.ovs_idl.transaction() as txn:
            txn.add(self.ovs_idl.add_br(self.br_name))
            txn.add(self.ovs_idl.add_port(self.br_name, self.port_name))
            txn.add(self.ovs_idl.iface_set_external_id(
                self.port_name, 'iface-id', port_iface_id))
            txn.add(self.ovs_idl.db_set(
                'Interface', self.port_name, ('type', 'internal')))

        exc = Exception('Port %s was not added to the bridge %s' %
                        (self.port_name, self.br_name))
        n_utils.wait_until_true(check_add_port_called, timeout=5,
                                exception=exc)

        self.ovs_idl.del_port(self.port_name).execute(check_error=True)
        exc = Exception('Port %s was not deleted from the bridge %s' %
                        (self.port_name, self.br_name))
        n_utils.wait_until_true(check_remove_egress_called, timeout=5,
                                exception=exc)


class QoSBandwidthLimitEventTestCase(base.TestOVNFunctionalBase):

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.net = self._make_network(self.fmt, 'n1', True)['network']
        res = self._create_subnet(self.fmt, self.net['id'], '10.0.0.0/24')
        self.subnet = self.deserialize(self.fmt, res)['subnet']
        res = self._create_port(self.fmt, self.net['id'])
        self.port = self.deserialize(self.fmt, res)['port']

    def test_qos_bw_limit_created_and_updated(self):
        def check_update_egress_called(rate):
            try:
                mock_agent[qos_hwol.EXT_NAME].update_egress.assert_has_calls(
                    [mock.call(port_id, rate, 0)])
                return True
            except AssertionError:
                return False

        mock_agent = mock.MagicMock(nb_idl=self.nb_api)
        events = [qos_hwol.QoSBandwidthLimitEvent(mock_agent)]
        agent_ovsdb.MonitorAgentOvnNbIdl(qos_hwol.NB_IDL_TABLES,
                                         events).start()
        lswitch_name = utils.ovn_name(self.net['id'])
        port_id = self.port['id']
        ovn_rule = {'switch': lswitch_name,
                    'priority': 1000,
                    'direction': 'from-lport',
                    'match': 'inport == ' + port_id,
                    'rate': 10000,
                    'external_ids': {ovn_const.OVN_PORT_EXT_ID_KEY: port_id}}
        self.nb_api.qos_add(**ovn_rule).execute(check_error=True)
        n_utils.wait_until_true(
            lambda: check_update_egress_called(ovn_rule['rate']), timeout=5)

        ovn_rule['rate'] = 15000
        self.nb_api.qos_add(**ovn_rule, may_exist=True).execute(
            check_error=True)
        n_utils.wait_until_true(
            lambda: check_update_egress_called(ovn_rule['rate']), timeout=5)


class QoSMinimumBandwidthEventTestCase(base.TestOVNFunctionalBase):

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.net = self._make_network(self.fmt, 'n1', True)['network']
        res = self._create_subnet(self.fmt, self.net['id'], '10.0.0.0/24')
        self.subnet = self.deserialize(self.fmt, res)['subnet']
        res = self._create_port(self.fmt, self.net['id'])
        self.port = self.deserialize(self.fmt, res)['port']

    def test_qos_min_bw_created_and_updated(self):
        def check_update_egress_called(max_kbps, min_kbps):
            try:
                mock_agent[qos_hwol.EXT_NAME].update_egress.assert_has_calls(
                    [mock.call(port_id, max_kbps, min_kbps)])
                return True
            except AssertionError:
                return False

        mock_agent = mock.MagicMock(nb_idl=self.nb_api)
        events = [qos_hwol.QoSMinimumBandwidthEvent(mock_agent)]
        agent_ovsdb.MonitorAgentOvnNbIdl(qos_hwol.NB_IDL_TABLES,
                                         events).start()
        port_id = self.port['id']
        min_kbps = 5000
        lsp = self.nb_api.lsp_get(port_id).execute(check_error=True)
        options = {ovn_const.LSP_OPTIONS_QOS_MIN_RATE: str(min_kbps)}
        self.nb_api.update_lswitch_qos_options(lsp, **options).execute(
            check_error=True)
        n_utils.wait_until_true(
            lambda: check_update_egress_called(0, min_kbps), timeout=5)


class PortBindingChassisCreatedEventTestCase(base.TestOVNFunctionalBase):

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.net = self._make_network(self.fmt, 'n1', True)['network']
        res = self._create_subnet(self.fmt, self.net['id'], '10.0.0.0/24')
        self.subnet = self.deserialize(self.fmt, res)['subnet']
        res = self._create_port(self.fmt, self.net['id'])
        self.port = self.deserialize(self.fmt, res)['port']

    @mock.patch.object(agent_ovsdb, 'get_ovs_port_name')
    @mock.patch.object(agent_ovsdb, 'get_port_qos')
    def test_port_binding_chassis_create_event(self, mock_get_port_qos,
                                               *args):
        def check_update_egress_called(max_kbps, min_kbps):
            try:
                mock_agent[qos_hwol.EXT_NAME].update_egress.assert_has_calls(
                    [mock.call(self.port['id'], max_kbps, min_kbps)])
                return True
            except AssertionError:
                return False

        max_kbps, min_kbps = 1000, 800
        mock_get_port_qos.return_value = max_kbps, min_kbps
        mock_agent = mock.MagicMock(nb_idl=self.nb_api)
        events = [qos_hwol.PortBindingChassisCreatedEvent(mock_agent)]
        chassis_name = self.add_fake_chassis('ovn-host-fake')
        mock_agent.chassis = chassis_name
        agent_ovsdb.MonitorAgentOvnSbIdl(qos_hwol.SB_IDL_TABLES, events,
                                         chassis=chassis_name).start()
        lsp_columns = {}
        lsp_name = self.port['id']
        ls_name = utils.ovn_name(self.net['id'])
        self.nb_api.create_lswitch_port(
            lsp_name, ls_name, **lsp_columns).execute(check_error=True)
        self.sb_api.lsp_bind(lsp_name, chassis_name).execute(check_error=True)
        n_utils.wait_until_true(
            lambda: check_update_egress_called(max_kbps, min_kbps), timeout=5)



# Copyright (c) 2017 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.definitions import data_plane_status as dps_lib
from neutron_lib.api.definitions import port as port_def
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron.plugins.ml2.extensions import data_plane_status
from neutron.tests.unit.plugins.ml2 import test_plugin


class DataPlaneStatusSML2ExtDriverTestCase(test_plugin.Ml2PluginV2TestCase):

    _extension_drivers = ['data_plane_status']

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers,
                              group='ml2')
        super(DataPlaneStatusSML2ExtDriverTestCase, self).setUp()
        self.plugin = directory.get_plugin()

    def test_extend_port_dict_no_data_plane_status(self):
        for db_data in ({'data_plane_status': None}, {}):
            response_data = {}
            session = mock.Mock()

            driver = data_plane_status.DataPlaneStatusExtensionDriver()
            driver.extend_port_dict(session, db_data, response_data)
            self.assertIsNone(response_data['data_plane_status'])

    def test_show_port_has_data_plane_status(self):
        with self.port() as port:
            req = self.new_show_request(port_def.COLLECTION_NAME,
                                        port['port']['id'],
                                        self.fmt)
            p = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertIsNone(p['port'][dps_lib.DATA_PLANE_STATUS])

    def test_port_update_data_plane_status(self):
        with self.port() as port:
            admin_ctx = context.get_admin_context()
            p = {'port': {dps_lib.DATA_PLANE_STATUS: constants.ACTIVE}}
            self.plugin.update_port(admin_ctx, port['port']['id'], p)
            req = self.new_show_request(
                port_def.COLLECTION_NAME, port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(res['port'][dps_lib.DATA_PLANE_STATUS],
                             constants.ACTIVE)



# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import fixture as fixture_config
from oslo_utils import uuidutils

from neutron.agent.ovn.agent import ovn_neutron_agent
from neutron.agent.ovn.agent import ovsdb as agent_ovsdb
from neutron.common import utils as n_utils
from neutron.tests.common import net_helpers
from neutron.tests.functional import base


TEST_EXTENSION = 'testing'


class TestOVNNeutronAgent(base.TestOVNFunctionalBase):

    OVN_BRIDGE = 'br-int'
    FAKE_CHASSIS_HOST = 'ovn-host-fake'

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.mock_chassis_name = mock.patch.object(
            agent_ovsdb, 'get_own_chassis_name').start()
        self.ovn_agent = self._start_ovn_neutron_agent()

    def _check_loaded_extensions(self, ovn_agent):
        loaded_ext = ovn_agent[TEST_EXTENSION]
        self.assertEqual('Fake OVN agent extension', loaded_ext.name)

    def _start_ovn_neutron_agent(self):
        conf = self.useFixture(fixture_config.Config()).conf
        conf.set_override('extensions', TEST_EXTENSION, group='agent')
        ovn_nb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('nb')
        conf.set_override('ovn_nb_connection', ovn_nb_db, group='ovn')
        ovn_sb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('sb')
        conf.set_override('ovn_sb_connection', ovn_sb_db, group='ovn')

        self.chassis_name = uuidutils.generate_uuid()
        self.mock_chassis_name.return_value = self.chassis_name

        agt = ovn_neutron_agent.OVNNeutronAgent(conf)
        agt.test_ovs_idl = []
        agt.test_ovn_sb_idl = []
        agt.test_ovn_nb_idl = []
        agt.start()
        self._check_loaded_extensions(agt)

        self.add_fake_chassis(self.FAKE_CHASSIS_HOST, name=self.chassis_name)

        self.addCleanup(agt.ext_manager_api.ovs_idl.ovsdb_connection.stop)
        if agt.ext_manager_api.sb_idl:
            self.addCleanup(agt.ext_manager_api.sb_idl.ovsdb_connection.stop)
        if agt.ext_manager_api.nb_idl:
            self.addCleanup(agt.ext_manager_api.nb_idl.ovsdb_connection.stop)
        return agt

    def test_ovs_and_ovs_events(self):
        # Test the OVS IDL is attending the provided events.
        bridge = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        exc = Exception('Bridge %s not added or not detected by '
                        'OVSInterfaceEvent' % bridge)
        n_utils.wait_until_true(
            lambda: bridge.br_name in self.ovn_agent.test_ovs_idl,
            timeout=10, exception=exc)

        # Test the OVN SB IDL is attending the provided events. The chassis is
        # created before the OVN SB IDL connection is created but the creation
        # event is received during the subscription.
        exc = Exception('Chassis %s not added or not detected by '
                        'OVNSBChassisEvent' % self.chassis_name)
        n_utils.wait_until_true(
            lambda: self.chassis_name in self.ovn_agent.test_ovn_sb_idl,
            timeout=10, exception=exc)

        # Test the OVN SN IDL is attending the provided events.
        lswitch_name = 'ovn-' + uuidutils.generate_uuid()
        self.nb_api.ls_add(lswitch_name).execute(check_error=True)
        exc = Exception('Logical Switch %s not added or not detected by ')
        n_utils.wait_until_true(
            lambda: lswitch_name in self.ovn_agent.test_ovn_nb_idl,
            timeout=10)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import orm


class RouterExtraAttributes(model_base.BASEV2):
    """Additional attributes for a Virtual Router."""

    # NOTE(armando-migliaccio): this model can be a good place to
    # add extension attributes to a Router model. Each case needs
    # to be individually examined, however 'distributed' and other
    # simple ones fit the pattern well.
    __tablename__ = "router_extra_attributes"
    router_id = sa.Column(sa.String(36),
                          sa.ForeignKey('routers.id', ondelete="CASCADE"),
                          primary_key=True)
    # Whether the router is a legacy (centralized) or a distributed one
    distributed = sa.Column(sa.Boolean, default=False,
                            server_default=sa.sql.false(),
                            nullable=False)
    # Whether the router is to be considered a 'service' router
    service_router = sa.Column(sa.Boolean, default=False,
                               server_default=sa.sql.false(),
                               nullable=False)
    ha = sa.Column(sa.Boolean, default=False,
                   server_default=sa.sql.false(),
                   nullable=False)
    ha_vr_id = sa.Column(sa.Integer())
    # Availability Zone support
    availability_zone_hints = sa.Column(sa.String(255))
    enable_default_route_ecmp = sa.Column(sa.Boolean, default=False,
                                          server_default=sa.sql.false(),
                                          nullable=False)
    enable_default_route_bfd = sa.Column(sa.Boolean, default=False,
                                         server_default=sa.sql.false(),
                                         nullable=False)

    router = orm.relationship(
        'Router', load_on_pending=True,
        backref=orm.backref("extra_attributes", lazy='joined',
                            uselist=False, cascade='delete'))
    revises_on_change = ('router', )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import constants as ovsdbapp_const

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as const
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.db import ovn_revision_numbers_db as db_rev
from neutron import manager
from neutron.objects import port_forwarding as port_forwarding_obj
from neutron.services.portforwarding import constants as pf_const

LOG = log.getLogger(__name__)


class OVNPortForwardingHandler(object):
    @staticmethod
    def _get_lb_protocol(pf_obj):
        return pf_const.LB_PROTOCOL_MAP[pf_obj.protocol]

    @staticmethod
    def lb_name(fip_id, proto, external_port=''):
        if external_port:
            external_port = '-%s' % external_port
        return "{}-{}-{}{}".format(
            pf_const.PORT_FORWARDING_PREFIX, fip_id, proto, external_port)

    @classmethod
    def lb_names(cls, fip_id):
        return [cls.lb_name(fip_id, proto)
                for proto in pf_const.LB_PROTOCOL_MAP.values()]

    @classmethod
    def _get_lb_attributes(cls, pf_obj, is_range=False):
        external_port = pf_obj.external_port if is_range else ''
        lb_name = cls.lb_name(pf_obj.floatingip_id,
                              cls._get_lb_protocol(pf_obj),
                              external_port)
        vip = "{}:{}".format(pf_obj.floating_ip_address, pf_obj.external_port)
        internal_ip = "{}:{}".format(pf_obj.internal_ip_address,
                                     pf_obj.internal_port)
        rtr_name = 'neutron-{}'.format(pf_obj.router_id)
        return lb_name, vip, [internal_ip], rtr_name

    def _get_lbs_and_ls(self, nb_ovn, payload):
        rtr_name = ovn_utils.ovn_name(payload.resource_id)
        ovn_lr = nb_ovn.get_lrouter(rtr_name)
        if ovn_lr:
            ext_id_key = ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY
            # Filter only lbs managed by port forwarding plugin
            lr_lbs = [lr for lr in ovn_lr.load_balancer
                      if lr.external_ids.get(ext_id_key) ==
                      pf_const.PORT_FORWARDING_PLUGIN]
            r_port = payload.metadata.get('port')

            if r_port:
                ls_name = ovn_utils.ovn_name(r_port['network_id'])
                ovn_ls = nb_ovn.get_lswitch(ls_name)
                ls_lbs = ovn_ls.load_balancer
                return lr_lbs, ls_lbs, ls_name
        return [], [], None

    def _add_lb_on_ls(self, ovn_txn, nb_ovn, payload):
        lr_lbs, ls_lbs, ls_name = self._get_lbs_and_ls(nb_ovn, payload)
        for lb in lr_lbs:
            if lb not in ls_lbs:
                ovn_txn.add(nb_ovn.ls_lb_add(ls_name, lb.name, may_exist=True))

    def _del_lb_on_ls(self, ovn_txn, nb_ovn, payload):
        lr_lbs, ls_lbs, ls_name = self._get_lbs_and_ls(nb_ovn, payload)
        for lb in lr_lbs:
            if lb in ls_lbs:
                ovn_txn.add(nb_ovn.ls_lb_del(ls_name, lb.name))

    def _port_forwarding_created(self, ovn_txn, nb_ovn, pf_obj,
                                 is_range=False):
        # Add vip to its corresponding load balancer. There can be multiple
        # vips, so load balancer may already be present.
        lb_name, vip, internal_ips, rtr_name = self._get_lb_attributes(
            pf_obj, is_range=is_range)
        external_ids = {
            ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                pf_const.PORT_FORWARDING_PLUGIN,
            ovn_const.OVN_FIP_EXT_ID_KEY: pf_obj.floatingip_id,
            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: rtr_name,
        }
        ovn_txn.add(
            nb_ovn.lb_add(lb_name, vip, internal_ips,
                          self._get_lb_protocol(pf_obj), may_exist=True,
                          external_ids=external_ids))
        # Ensure logical router has load balancer configured.
        ovn_txn.add(nb_ovn.lr_lb_add(rtr_name, lb_name, may_exist=True))
        # Ensure logical switches on logical router have load balancer
        # configured, can be removed this handling if in future ovn
        # supports auto handling of lbs on ls rhbz#2043543
        ovn_lr = nb_ovn.get_lrouter(rtr_name)
        if ovn_lr:
            ext_id_key = ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY
            ovn_lss = [port.external_ids.get(ext_id_key)
                       for port in ovn_lr.ports
                       if port.external_ids.get(ext_id_key) and
                       not port.gateway_chassis]
            for ls_name in ovn_lss:
                try:
                    ovn_txn.add(nb_ovn.ls_lb_add(ls_name, lb_name,
                                                 may_exist=True))
                except idlutils.RowNotFound:
                    # If one or more logical_switches are deleted
                    # log warning for those and continue with the rest.
                    LOG.warning("Addition of Load Balancer %s to Logical "
                                "Switch %s failed as it is not found",
                                lb_name, ls_name)

    def port_forwarding_created(self, ovn_txn, nb_ovn, pf_obj):
        pf_objs = pf_obj.unroll_port_ranges()
        is_range = len(pf_objs) > 1
        for pf_obj in pf_objs:
            LOG.info("CREATE for port-forwarding %s vip %s:%s to %s:%s",
                     pf_obj.protocol,
                     pf_obj.floating_ip_address, pf_obj.external_port,
                     pf_obj.internal_ip_address, pf_obj.internal_port)
            self._port_forwarding_created(ovn_txn, nb_ovn, pf_obj,
                                          is_range=is_range)

    def port_forwarding_updated(self, ovn_txn, nb_ovn, pf_obj, orig_pf_obj):
        orig_pf_objs = orig_pf_obj.unroll_port_ranges()
        is_range = len(orig_pf_objs) > 1
        for orig_pf_obj in orig_pf_objs:
            self._port_forwarding_deleted(ovn_txn, nb_ovn, orig_pf_obj,
                                          is_range=is_range)
        pf_objs = pf_obj.unroll_port_ranges()
        is_range = len(pf_objs) > 1
        for pf_obj in pf_objs:
            LOG.info("UPDATE for port-forwarding %s vip %s:%s to %s:%s",
                     pf_obj.protocol,
                     pf_obj.floating_ip_address, pf_obj.external_port,
                     pf_obj.internal_ip_address, pf_obj.internal_port)
            self._port_forwarding_created(ovn_txn, nb_ovn, pf_obj,
                                          is_range=is_range)

    def _port_forwarding_deleted(self, ovn_txn, nb_ovn, pf_obj,
                                 is_range=False):
        # NOTE: load balancer instance is expected to be removed by api once
        #       last vip is removed.
        #       Since router has weak ref to the lb, that gets taken care
        #       automatically, but that it is not best practice to rely on
        #       that. Unfortunately, we would need to add extra logic that
        #       ensures that the lr_lb_del is invoked only after the last
        #       vip was removed. So...
        # TODO(flaviof): see about enhancing lb_del so that removal of lb
        # can optionally take a logical router, which explicitly dissociates
        # router from removed lb.
        pf_objs = pf_obj.unroll_port_ranges()
        is_range = is_range or len(pf_objs) > 1
        for pf_obj in pf_objs:
            lb_name, vip, _internal_ips, _rtr = self._get_lb_attributes(
                pf_obj, is_range=is_range)
            ovn_txn.add(nb_ovn.lb_del(lb_name, vip, if_exists=True))

    def port_forwarding_deleted(self, ovn_txn, nb_ovn, pf_obj):
        pf_objs = pf_obj.unroll_port_ranges()
        is_range = len(pf_objs) > 1
        for pf_obj in pf_objs:
            LOG.info("DELETE for port-forwarding %s vip %s:%s to %s:%s",
                     pf_obj.protocol,
                     pf_obj.floating_ip_address, pf_obj.external_port,
                     pf_obj.internal_ip_address, pf_obj.internal_port)
            self._port_forwarding_deleted(ovn_txn, nb_ovn, pf_obj,
                                          is_range=is_range)


@registry.has_registry_receivers
class OVNPortForwarding(object):

    def __init__(self, l3_plugin):
        self._l3_plugin = l3_plugin
        self._pf_plugin_property = None
        self._handler = OVNPortForwardingHandler()

    @property
    def _pf_plugin(self):
        if self._pf_plugin_property is None:
            self._pf_plugin_property = directory.get_plugin(
                plugin_constants.PORTFORWARDING)
            if not self._pf_plugin_property:
                self._pf_plugin_property = (
                    manager.NeutronManager.load_class_for_provider(
                        'neutron.service_plugins', 'port_forwarding')())
        return self._pf_plugin_property

    def _get_pf_objs(self, context, fip_id):
        pf_dicts = self._pf_plugin.get_floatingip_port_forwardings(
            context, fip_id)
        return [port_forwarding_obj.PortForwarding(context=context, **pf_dict)
                for pf_dict in pf_dicts]

    def _get_fip_objs(self, context, payload):
        floatingip_ids = set()
        for fip in payload.states:
            floatingip_ids.add(fip.floatingip_id)
        return {fip_id: self._l3_plugin.get_floatingip(context, fip_id)
                for fip_id in floatingip_ids}

    def _add_check_rev(self, ovn_txn, ovn_nb, fip_id, fip_obj):
        """Updating revision number of OVN lb entries based on floatingip id

           A single floating ip maps to 1 or 2 OVN load balancer entries,
           because while multiple vips can exist in a single OVN LB row,
           they represent one protocol. So, to handle all port forwardings
           for a given floating ip, OVN will have up to two LB entries: one
           for udp and one for tcp. These 2 LB entries are expected to have
           the same revision number, in sync with the revision of the floating
           ip. And that is set via this function.
        """
        check_rev_tuples = []
        for lb_name in self._handler.lb_names(fip_id):
            check_rev_cmd = ovn_nb.check_revision_number(
                lb_name, fip_obj, ovn_const.TYPE_FLOATINGIPS, if_exists=True)
            ovn_txn.add(check_rev_cmd)
            check_rev_tuples.append((check_rev_cmd, fip_obj))
        return check_rev_tuples

    def _do_db_rev_bump_revision(self, context, check_rev_tuples):
        if not all(check_rev_cmd.result == ovn_const.TXN_COMMITTED
                   for check_rev_cmd, _fip_obj in check_rev_tuples):
            return
        for _check_rev_cmd, fip_obj in check_rev_tuples:
            db_rev.bump_revision(context, fip_obj, ovn_const.TYPE_FLOATINGIPS)

    def _handle_notification(self, _resource, event_type, _pf_plugin, payload):
        if not payload:
            return
        context = payload.context
        ovn_nb = self._l3_plugin._nb_ovn
        with ovn_nb.transaction(check_error=True) as ovn_txn:
            if event_type == events.AFTER_CREATE:
                self._handler.port_forwarding_created(ovn_txn, ovn_nb,
                                                      payload.latest_state)
                self._l3_plugin.update_floatingip_status(
                    context, payload.latest_state.floatingip_id,
                    const.FLOATINGIP_STATUS_ACTIVE)
            elif event_type == events.AFTER_UPDATE:
                self._handler.port_forwarding_updated(
                    ovn_txn, ovn_nb,
                    payload.latest_state, payload.states[0])
            elif event_type == events.AFTER_DELETE:
                pfs = _pf_plugin.get_floatingip_port_forwardings(
                    context, payload.states[0].floatingip_id)
                self._handler.port_forwarding_deleted(ovn_txn, ovn_nb,
                                                      payload.states[0])
                if not pfs:
                    self._l3_plugin.update_floatingip_status(
                        context, payload.states[0].floatingip_id,
                        const.FLOATINGIP_STATUS_DOWN)

            # Collect the revision numbers of all floating ips visited and
            # update the corresponding load balancer entries affected.
            # Note that there may be 2 entries for a given floatingip_id;
            # one for each protocol.
            fip_objs = self._get_fip_objs(context, payload)
            if not fip_objs:
                return
            for floatingip_id, fip_obj in fip_objs.items():
                check_rev_tuples = self._add_check_rev(
                    ovn_txn, ovn_nb, floatingip_id, fip_obj)
        # Update revision of affected floating ips. Note that even in
        # cases where port forwarding is deleted, floating ip remains.
        self._do_db_rev_bump_revision(context, check_rev_tuples)

    def _maintenance_create_update(self, context, fip_id):
        # NOTE: Since the maintenance callback is not granular to the level
        #       of the affected pfs AND the fact that pfs are all vips
        #       in a load balancer entry, it is cheap enough to simply rebuild.
        pf_objs = self._get_pf_objs(context, fip_id)
        LOG.debug("Maintenance port forwarding under fip %s : %s",
                  fip_id, pf_objs)
        ovn_nb = self._l3_plugin._nb_ovn
        with ovn_nb.transaction(check_error=True) as ovn_txn:
            for lb_name in self._handler.lb_names(fip_id):
                ovn_txn.add(ovn_nb.lb_del(lb_name, vip=None, if_exists=True))
            for pf_obj in pf_objs:
                self._handler.port_forwarding_created(
                    ovn_txn, ovn_nb, pf_obj)
            fip_obj = self._l3_plugin.get_floatingip(context, fip_id)
            check_rev_tuples = self._add_check_rev(
                ovn_txn, ovn_nb, fip_id, fip_obj)
            self._do_db_rev_bump_revision(context, check_rev_tuples)

    def maintenance_create(self, context, floatingip):
        fip_id = floatingip['id']
        LOG.info("Maintenance CREATE port-forwarding entries under fip %s",
                 fip_id)
        self._maintenance_create_update(context, fip_id)

    def maintenance_update(self, context, floatingip):
        fip_id = floatingip['id']
        LOG.info("Maintenance UPDATE port-forwarding entries under fip %s",
                 fip_id)
        self._maintenance_create_update(context, fip_id)

    def maintenance_delete(self, _context, fip_id):
        LOG.info("Maintenance DELETE port-forwarding entries under fip %s",
                 fip_id)
        ovn_nb = self._l3_plugin._nb_ovn
        with ovn_nb.transaction(check_error=True) as ovn_txn:
            for lb_name in self._handler.lb_names(fip_id):
                ovn_txn.add(ovn_nb.lb_del(lb_name, vip=None, if_exists=True))

    def db_sync_create_or_update(self, context, fip_id, ovn_txn):
        LOG.info("db_sync UPDATE entries under fip %s", fip_id)
        # NOTE: Since the db_sync callback is not granular to the level
        #       of the affected pfs AND the fact that pfs are all vips
        #       in a load balancer entry, it is cheap enough to simply rebuild.
        ovn_nb = self._l3_plugin._nb_ovn
        pf_objs = self._get_pf_objs(context, fip_id)
        LOG.debug("Db sync port forwarding under fip %s : %s", fip_id, pf_objs)
        for lb_name in self._handler.lb_names(fip_id):
            ovn_txn.add(ovn_nb.lb_del(lb_name, vip=None, if_exists=True))
        for pf_obj in pf_objs:
            self._handler.port_forwarding_created(ovn_txn, ovn_nb, pf_obj)
        fip_obj = self._l3_plugin.get_floatingip(context, fip_id)
        self._add_check_rev(ovn_txn, ovn_nb, fip_id, fip_obj)

    def db_sync_delete(self, context, fip_id, ovn_txn):
        LOG.info("db_sync DELETE entries under fip %s", fip_id)
        ovn_nb = self._l3_plugin._nb_ovn
        for lb_name in self._handler.lb_names(fip_id):
            ovn_txn.add(ovn_nb.lb_del(lb_name, vip=None, if_exists=True))

    def _handle_lb_on_ls(self, _resource, event_type, _pf_plugin, payload):
        if not payload:
            return
        ovn_nb = self._l3_plugin._nb_ovn
        with ovn_nb.transaction(check_error=True) as ovn_txn:
            if event_type == events.AFTER_CREATE:
                self._handler._add_lb_on_ls(ovn_txn, ovn_nb,
                                            payload)
            if event_type == events.AFTER_DELETE:
                self._handler._del_lb_on_ls(ovn_txn, ovn_nb,
                                            payload)

    @staticmethod
    def ovn_lb_protocol(pf_protocol):
        return pf_const.LB_PROTOCOL_MAP.get(
            pf_protocol, ovsdbapp_const.PROTO_TCP)

    @registry.receives(pf_const.PORT_FORWARDING_PLUGIN, [events.AFTER_INIT])
    def register(self, resource, event, trigger, payload=None):
        for event_type in (events.AFTER_CREATE, events.AFTER_UPDATE,
                           events.AFTER_DELETE):
            registry.subscribe(self._handle_notification,
                               pf_const.PORT_FORWARDING, event_type)
            if event_type in (events.AFTER_CREATE, events.AFTER_DELETE):
                registry.subscribe(self._handle_lb_on_ls,
                                   resources.ROUTER_INTERFACE, event_type)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.callbacks import exceptions

from neutron.api.rpc.callbacks import exceptions as rpc_exc
from neutron.api.rpc.callbacks import resource_manager
from neutron.tests.unit.services.qos import base

IS_VALID_RESOURCE_TYPE = (
    'neutron.api.rpc.callbacks.resources.is_valid_resource_type')


class ResourceCallbacksManagerTestCaseMixin(object):

    def test_register_fails_on_invalid_type(self):
        self.assertRaises(
            exceptions.Invalid,
            self.mgr.register, lambda: None, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_clear_unregisters_all_callbacks(self, *mocks):
        self.mgr.register(lambda: None, 'TYPE1')
        self.mgr.register(lambda: None, 'TYPE2')
        self.mgr.clear()
        self.assertEqual([], self.mgr.get_subscribed_types())

    def test_unregister_fails_on_invalid_type(self):
        self.assertRaises(
            exceptions.Invalid,
            self.mgr.unregister, lambda: None, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_unregister_fails_on_unregistered_callback(self, *mocks):
        self.assertRaises(
            rpc_exc.CallbackNotFound,
            self.mgr.unregister, lambda: None, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_unregister_unregisters_callback(self, *mocks):
        callback = lambda: None
        self.mgr.register(callback, 'TYPE')
        self.mgr.unregister(callback, 'TYPE')
        self.assertEqual([], self.mgr.get_subscribed_types())

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test___init___does_not_reset_callbacks(self, *mocks):
        callback = lambda: None
        self.mgr.register(callback, 'TYPE')
        resource_manager.ProducerResourceCallbacksManager()
        self.assertEqual(['TYPE'], self.mgr.get_subscribed_types())


class ProducerResourceCallbacksManagerTestCase(
        base.BaseQosTestCase, ResourceCallbacksManagerTestCaseMixin):

    def setUp(self):
        super(ProducerResourceCallbacksManagerTestCase, self).setUp()
        self.mgr = self.prod_mgr

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_register_registers_callback(self, *mocks):
        callback = lambda: None
        self.mgr.register(callback, 'TYPE')
        self.assertEqual(callback, self.mgr.get_callback('TYPE'))

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_register_fails_on_multiple_calls(self, *mocks):
        self.mgr.register(lambda: None, 'TYPE')
        self.assertRaises(
            rpc_exc.CallbacksMaxLimitReached,
            self.mgr.register, lambda: None, 'TYPE')

    def test_get_callback_fails_on_invalid_type(self):
        self.assertRaises(
            exceptions.Invalid,
            self.mgr.get_callback, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_get_callback_fails_on_unregistered_callback(
            self, *mocks):
        self.assertRaises(
            rpc_exc.CallbackNotFound,
            self.mgr.get_callback, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_get_callback_returns_proper_callback(self, *mocks):
        callback1 = lambda: None
        callback2 = lambda: None
        self.mgr.register(callback1, 'TYPE1')
        self.mgr.register(callback2, 'TYPE2')
        self.assertEqual(callback1, self.mgr.get_callback('TYPE1'))
        self.assertEqual(callback2, self.mgr.get_callback('TYPE2'))


class ConsumerResourceCallbacksManagerTestCase(
        base.BaseQosTestCase, ResourceCallbacksManagerTestCaseMixin):

    def setUp(self):
        super(ConsumerResourceCallbacksManagerTestCase, self).setUp()
        self.mgr = self.cons_mgr

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_register_registers_callback(self, *mocks):
        callback = lambda: None
        self.mgr.register(callback, 'TYPE')
        self.assertEqual({callback}, self.mgr.get_callbacks('TYPE'))

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_register_succeeds_on_multiple_calls(self, *mocks):
        callback1 = lambda: None
        callback2 = lambda: None
        self.mgr.register(callback1, 'TYPE')
        self.mgr.register(callback2, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_get_callbacks_fails_on_unregistered_callback(self, *mocks):
        self.assertRaises(
            rpc_exc.CallbackNotFound,
            self.mgr.get_callbacks, 'TYPE')

    @mock.patch(IS_VALID_RESOURCE_TYPE, return_value=True)
    def test_get_callbacks_returns_proper_callbacks(self, *mocks):
        callback1 = lambda: None
        callback2 = lambda: None
        self.mgr.register(callback1, 'TYPE1')
        self.mgr.register(callback2, 'TYPE2')
        self.assertEqual(set([callback1]), self.mgr.get_callbacks('TYPE1'))
        self.assertEqual(set([callback2]), self.mgr.get_callbacks('TYPE2'))



# Copyright 2023 Ericsson Software Technology
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron.objects.port.extensions import port_hints
from neutron.tests.unit.objects import test_base as obj_test_base
from neutron.tests.unit import testlib_api


class PortHintsIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = port_hints.PortHints


class PortHintsDbObjectTestCase(
        obj_test_base.BaseDbObjectTestCase, testlib_api.SqlTestCase):

    _test_class = port_hints.PortHints

    def setUp(self):
        super(PortHintsDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'port_id': lambda: self._create_test_port_id()})



#!/usr/bin/env python3
# Copyright (c) 2019 Ericsson
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import sys
import urllib
import uuid
from wsgiref import simple_server as wsgi_simple_server

from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
from oslo_serialization import jsonutils

from neutron.common import config as common_config

LOG = logging.getLogger(__name__)


PortType = types.Integer(1, 65535)
placement_opts = [
    cfg.Opt('placement_port', type=PortType)
]
cfg.CONF.register_opts(placement_opts)


class FakePlacement(object):

    rp_template = {
        "uuid": None,
        "generation": 0,
        "parent_provider_uuid": None,
        "name": None,
    }

    def __init__(self):
        self.host_rp_uuid = str(uuid.uuid4())
        host_rp = copy.deepcopy(self.rp_template)
        host_rp['uuid'] = self.host_rp_uuid
        self.resource_providers = {
            self.host_rp_uuid: host_rp
        }

    def get_resource_providers(self, **kwargs):
        id = kwargs.get('id', None)
        if not id:
            return jsonutils.dumps(
                {
                    'resource_providers':
                        [self.resource_providers[self.host_rp_uuid]]
                })
        else:
            return jsonutils.dumps(self.resource_providers[id])

    def put_traits(self, **kwargs):
        # Return empty sting otherwise wsgiref goes mad
        return ''

    def put_resource_providers(self, **kwargs):
        id = kwargs.get('id', None)
        req_body = kwargs.get('body', None)
        if id:
            rp_dict = copy.deepcopy(self.rp_template)
            rp_dict['uuid'] = id
            rp_dict['parent_provider_uuid'] = req_body['parent_provider_uuid']
            rp_dict['name'] = req_body['name']
            self.resource_providers[rp_dict['uuid']] = rp_dict
            return jsonutils.dumps(rp_dict)

    def put_resource_providers_traits(self, **kwargs):
        resp = kwargs['body']
        resp['resource_provider_generation'] += 1
        return jsonutils.dumps(resp)

    def put_resource_providers_inventories(self, **kwargs):
        resp = kwargs['body']
        resp['resource_provider_generation'] += 1
        return jsonutils.dumps(resp)

    def build_method_name(self, action, path_info):
        path_info = urllib.parse.urlparse(path_info).path
        path_info_list = path_info.strip('/').split('/')
        method_name = action.lower()
        for path_chunk in path_info_list:
            if any(s in path_chunk for s in ['placement', 'CUSTOM']):
                continue
            # If there is uuid in the path, that should be thrown out
            try:
                uuid.UUID(path_chunk)
            except ValueError:
                method_name += '_' + path_chunk
        return method_name

    def wsgi_app(self, env, start_response):
        response_headers = [('Content-Type', 'application/json')]
        http_status = '200 OK'

        # Poor men's routing
        meth_name = self.build_method_name(env['REQUEST_METHOD'],
                                           env['PATH_INFO'])
        params = {}
        # Fetch params from url
        try:
            params['id'] = env['PATH_INFO'].split('/')[3]
        except IndexError:
            pass
        # Fetch body
        try:
            request_body_size = int(env.get('CONTENT_LENGTH', 0))
            if request_body_size > 0:
                req_body = env['wsgi.input'].read(request_body_size)
                params['body'] = jsonutils.loads(req_body.decode('utf-8'))
        except ValueError:
            pass
        LOG.debug('Request on %s (%s) with body: %s',
                  env['PATH_INFO'], env['REQUEST_METHOD'], params)
        response = getattr(self, meth_name)(**params)
        LOG.debug('Response from %s: %s', meth_name, response)

        response = response.encode('utf-8')
        start_response(http_status, response_headers)
        return [response]


if __name__ == "__main__":
    common_config.register_common_config_options()
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    placement_port = cfg.CONF.placement_port

    LOG.info("Placement fixture started on port: %s", placement_port)

    mock_placement = FakePlacement()
    wsgi_simple_server.make_server(
        '', placement_port, mock_placement.wsgi_app).serve_forever()



# Copyright (c) 2015 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from alembic.operations import ops
from alembic.util import Dispatcher
from alembic.util import rev_id as new_rev_id

from neutron._i18n import _
from neutron.db.migration import cli

_ec_dispatcher = Dispatcher()


def process_revision_directives(context, revision, directives):
    directives[:] = list(_assign_directives(context, directives))


def _assign_directives(context, directives, phase=None):
    for directive in directives:
        decider = _ec_dispatcher.dispatch(directive)
        if phase is None:
            phases = cli.MIGRATION_BRANCHES
        else:
            phases = (phase,)
        for phase in phases:
            decided = decider(context, directive, phase)
            if decided:
                yield decided


@_ec_dispatcher.dispatch_for(ops.MigrationScript)
def _migration_script_ops(context, directive, phase):
    """Generate a new ops.MigrationScript() for a given phase.

    E.g. given an ops.MigrationScript() directive from a vanilla autogenerate
    and an expand/contract phase name, produce a new ops.MigrationScript()
    which contains only those sub-directives appropriate to "expand" or
    "contract".  Also ensure that the branch directory exists and that
    the correct branch labels/depends_on/head revision are set up.

    """
    version_path = cli._get_version_branch_path(
        context.config, release=cli.CURRENT_RELEASE, branch=phase)
    autogen_kwargs = {}
    cli._check_bootstrap_new_branch(phase, version_path, autogen_kwargs)

    op = ops.MigrationScript(
        new_rev_id(),
        ops.UpgradeOps(ops=list(
            _assign_directives(context, directive.upgrade_ops.ops, phase)
        )),
        ops.DowngradeOps(ops=[]),
        message=directive.message,
        **autogen_kwargs
    )

    if not op.upgrade_ops.is_empty():
        return op


@_ec_dispatcher.dispatch_for(ops.AddConstraintOp)
@_ec_dispatcher.dispatch_for(ops.CreateIndexOp)
@_ec_dispatcher.dispatch_for(ops.CreateTableOp)
@_ec_dispatcher.dispatch_for(ops.AddColumnOp)
def _expands(context, directive, phase):
    if phase == 'expand':
        return directive
    else:
        return None


@_ec_dispatcher.dispatch_for(ops.DropConstraintOp)
@_ec_dispatcher.dispatch_for(ops.DropIndexOp)
@_ec_dispatcher.dispatch_for(ops.DropTableOp)
@_ec_dispatcher.dispatch_for(ops.DropColumnOp)
def _contracts(context, directive, phase):
    if phase == 'contract':
        return directive
    else:
        return None


@_ec_dispatcher.dispatch_for(ops.AlterColumnOp)
def _alter_column(context, directive, phase):
    is_expand = phase == 'expand'

    if is_expand and directive.modify_nullable is True:
        return directive
    elif not is_expand and directive.modify_nullable is False:
        return directive
    else:
        raise NotImplementedError(
            _("Don't know if operation is an expand or "
              "contract at the moment: %s") % directive)


@_ec_dispatcher.dispatch_for(ops.ModifyTableOps)
def _modify_table_ops(context, directive, phase):
    op = ops.ModifyTableOps(
        directive.table_name,
        ops=list(_assign_directives(context, directive.ops, phase)),
        schema=directive.schema)
    if not op.is_empty():
        return op



# Copyright (c) 2018 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import rpc as n_rpc
from oslo_concurrency import lockutils
from oslo_log import log as logging

from neutron.api.rpc.callbacks.consumer import registry
from neutron.api.rpc.callbacks import events
from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.handlers import resources_rpc
from neutron.conf.services import logging as log_cfg
from neutron import manager

LOG = logging.getLogger(__name__)

log_cfg.register_log_driver_opts()


class L3LoggingExtensionBase(object):
    """Base class for l3 logging extension like

    SNATLogExtension, FWaaSV2LogExtension
    """

    SUPPORTED_RESOURCE_TYPES = [resources.LOGGING_RESOURCE]

    def consume_api(self, agent_api):
        self.agent_api = agent_api

    def _load_driver_cls(self, namespace, driver_name):
        return manager.NeutronManager.load_class_for_provider(
            namespace, driver_name)

    def _register_rpc_consumers(self):
        registry.register(
            self._handle_notification, resources.LOGGING_RESOURCE)
        self._connection = n_rpc.Connection()
        endpoints = [resources_rpc.ResourcesPushRpcCallback()]
        topic = resources_rpc.resource_type_versioned_topic(
            resources.LOGGING_RESOURCE)
        self._connection.create_consumer(topic, endpoints, fanout=True)
        self._connection.consume_in_threads()

    def _get_router_info(self, router_id):
        router_info = self.agent_api.get_router_info(router_id)
        if router_info:
            return router_info
        LOG.debug("Router %s is not managed by this agent. "
                  "It was possibly deleted concurrently.",
                  router_id)

    @lockutils.synchronized('log')
    def _handle_notification(self, context, resource_type,
                             log_resources, event_type):
        with self.log_driver.defer_apply():
            if event_type == events.UPDATED:
                self._update_logging(context, log_resources)
            elif event_type == events.CREATED:
                self.log_driver.start_logging(
                    context, log_resources=log_resources)
            elif event_type == events.DELETED:
                self.log_driver.stop_logging(
                    context, log_resources=log_resources)

    def _update_logging(self, context, log_resources):
        enables = []
        disables = []
        for log_resource in log_resources:
            if log_resource.enabled:
                enables.append(log_resource)
            else:
                disables.append(log_resource)
        if enables:
            self.log_driver.start_logging(context, log_resources=enables)
        if disables:
            self.log_driver.stop_logging(context, log_resources=disables)

    def _process_update_router(self, context, router):
        router_info = self._get_router_info(router['id'])
        if router_info:
            self.log_driver.start_logging(context, router_info=router_info)

    @lockutils.synchronized('log-port')
    def add_router(self, context, data):
        self._process_update_router(context, data)

    @lockutils.synchronized('log-port')
    def update_router(self, context, data):
        self._process_update_router(context, data)

    def delete_router(self, context, data):
        self.log_driver.stop_logging(context, router_info=data)

    def ha_state_change(self, context, data):
        pass

    def update_network(self, context, data):
        pass
